/***********************************************************************/
/* pfDebug.c++                                                         */
/*                                                                     */
/* Routines for aiding in debuging Performer applications.             */
/* In particular, for examining and modifying scenegraphs.             */
/*                                                                     */
/* Also included are the curses routines used to display the scengraph.*/
/*                                                                     */
/* USAGE (coding):                                                     */
/*	Three functions need to be declared and used:                  */
/*		int	pfdebug_initialize_display(char *tty);         */
/*		void	pfdebug_terminate_display();                   */
/*		void	pfdebug_display_sg(pfScene *scene);            */
/*	Call pfdebug_initialize_display() at the beginning of the      */
/*	program, pfdebug_terminate_display() just before quitting, and */
/*	call pfdebug_display_sg() somewhere in the update loop.        */
/*                                                                     */
/*	For example, in a pfCAVE program, call ..._initialize_display()*/
/*	after pfCAVEInitChannels() -- it may work sooner, but this is  */
/*	fine.  Call pfdebug_terminate_display() before calling         */
/*	pfCAVEHalt().  And call pfdebug_display_sg() inside the        */
/*	application loop, prior to calling pfSync().                   */
/*                                                                     */
/*	The 'tty' string passed to pfdebug_initialize_display() should */
/*	be the file name associated with a terminal screen.  This will */
/*	be the terminal on which the scenegraph debugging information  */
/*	will be presented.  If the given file is unusable the function */
/*	will return 0, and the other two functions should not be used. */
/*                                                                     */
/* USAGE (using):                                                      */
/*	The pfDebug.c++ routines provide a variety of functionality.   */
/*	The simplest is simply to print a simple text description of   */
/*	the scenegraph (or portion thereof).  The next major features  */
/*	allow the user to view details on any node in the scenegraph   */
/*	by stepping through the graph.  Finally, the user can hide,    */
/*	unhide, cut, paste and shift nodes within the scenegraph.      */
/*	Note that when these modification operations are applied to    */
/*	a group node, then all the children of that node are likewise  */
/*	affected.                                                      */
/*                                                                     */
/*	The programmer can modify what keys are mapped to all the      */
/*	functions, but the default mapping is:                         */
/*	    SceneGraph traversal:                                      */
/*		F10 - move down a layer                                */
/*		F11 - move up a layer                                  */
/*		F9 - move left a child (sibling)                       */
/*		F12 - move right a child (sibling)                     */
/*                                                                     */
/*		Page Down - advance down the list of children by 10    */
/*		Page Up - move up the list of children by 10           */
/*		'+' - advance down the list of children by 1           */
/*		'-' - move up the list of children by 1                */
/*		Home - move to the first child                         */
/*		End - move to a page with the last child               */
/*                                                                     */
/*	    SceneGraph modification:                                   */
/*		F8 - hide/unhide the current node (toggle)             */
/*		Backspace - cut the current node from the scenegraph   */
/*		` - paste the previously cut node into the scenegraph  */
/*			(to just before current node)                  */
/*		F5 - move current node to the left among its siblings  */
/*		F6 - move current node to the right among its siblings */
/*                                                                     */
/*	    Misc:                                                      */
/*		Print Screen - print scenegraph from the current node  */
/*		F4 - print the detailed node info to stdout            */
/*                                                                     */
/*	While pfdebug.c++ does not provide the code for selecting the  */
/*	tty of the terminal on which to present the information, the   */
/*	way I, the code's author, have chosen for example applications */
/*	is to use the DBGTTY environment variable.  Thus, one can open */
/*	two terminal windows, one in which to run the program, and     */
/*	another larger window in which to view the debug display.      */
/*	Then, get the tty file for the large window by typing 'tty' in */
/*	it.  In the program window, set the DBGTTY environment variable*/
/*	to the tty value given.                                        */
/*                                                                     */
/* TODO:                                                               */
/*	- move the contents of the top window down to the bottom       */
/*	- add the ability to modify some nodes (eg. esp. pfSwitch)     */
/*	- when hiding a node, change the name of the 'empty' node      */
/*	- allow all nodes to independantly be hidden                   */
/*	- allow a simple (ascii) scenegraph rendering (perhaps in a    */
/*	  separate curses window -- though that might be difficult).   */
/*	- complete the vertex/tex index, etc info in the pfNodeString()*/
/*	- indexed geoset attributes are currently not handled          */
/*	- implement a geostate browser                                 */
/*	- add some geostate information to the pfNodeString() text     */
/*	- allow user to activate pfHighlight'ing on a geoset           */
/*	- function (and key command) to close and reopen curses in     */
/*	  order to readjust to a resized window.                       */
/*	- implement a main() program that allows one to specify a      */
/*	  scenegraph on the command line, and then traverse it without */
/*	  doing any graphics.                                          */
/*	- determine if there is a reasonable way to highlight text     */
/*	  in curses (ie. by embeding codes right in the text, since    */
/*	  the text creation is separate from the text rendering.)      */
/*                                                                     */
/* NOTES:                                                              */
/*	- A note on style: this source file does not represent my      */
/*	  preferred/recommended style of using separate files for      */
/*	  semantically separate functionality.  (In this case, I would */
/*	  normally have files for the curses routines, scenegraph text */
/*	  conversion routines, and CAVE library interface routines.)   */
/*	  Typically, I would either include the object files from these*/
/*	  three files directly, or put them together in a library.     */
/*	  However, for the sake of VR programming example code, I think*/
/*	  it's better to put it all in one file that the student can   */
/*	  keep easily in one place, and view the source.               */
/*                                                                     */
/* Copyright 2000, William R. Sherman, all rights reserved             */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>



/*****************************************************************************/
/*** ===================================================================== ***/
/*****************************************************************************/

/* Curses-library routines to display some debuging material to a text */
/*   screen.  Specifically these functions are used for a text view of */
/*   a performer scene graph.                                          */
/* See manual section "(3x) Subroutines" for reference information.    */

#include <curses.h>


static	WINDOW	*debug1_win,
		*debug2_win,
		*debug3_win;
	SCREEN	*scr = NULL;


/**********************************************************************/
static void clear_screen()
{
	wclear(stdscr);
	wrefresh(stdscr);
}


/**********************************************************************/
static void redisplay_screen()
{
	touchwin(debug1_win);
	wrefresh(debug1_win);
	touchwin(debug2_win);
	wrefresh(debug2_win);
	touchwin(debug3_win);
	wrefresh(debug3_win);
}


/**********************************************************************/
static void display_debug1(char *text)
{
static	char	*win_title = "Debug1 Window";
	char	save_char = '\0';

	if (strlen(text) > 4095) {
		save_char = text[4095];
		text[4095] = '\0';
	}

	werase(debug1_win);
	mvwprintw(debug1_win, 1, 2, text);
	box(debug1_win, '*', '*');
	mvwprintw(debug1_win, 0, (22 - strlen(win_title)/2), " %s ", win_title);
	wrefresh(debug1_win);

	/* only restore if saved because text memory might be less than 4096 bytes */
	if (save_char != '\0')
		text[4095] = save_char;
}


/**********************************************************************/
static void display_debug1_line(int line, char *text)
{
static	char	*win_title = "Debug1 Window";
	char	save_char = '\0';

	if (strlen(text) > 4095) {
		save_char = text[4095];
		text[4095] = '\0';
	}

#if 0
	werase(debug1_win);
#endif
	mvwprintw(debug1_win, line, 2, text);
	wclrtoeol(debug1_win);
	box(debug1_win, '*', '*');
	mvwprintw(debug1_win, 0, (22 - strlen(win_title)/2), " %s ", win_title);
	wrefresh(debug1_win);

	/* only restore if saved because text memory might be less than 4096 bytes */
	if (save_char != '\0')
		text[4095] = save_char;
}


/**********************************************************************/
static void display_debug2(char *text)
{
static	char	*win_title = "Debug2 Window";
	char	save_char = '\0';

	if (strlen(text) > 4095) {
		save_char = text[4095];
		text[4095] = '\0';
	}

	werase(debug2_win);
	mvwprintw(debug2_win, 2, 2, text);
	box(debug2_win, '*', '*');
	mvwprintw(debug2_win, 0, (22 - strlen(win_title)/2), " %s ", win_title);
	wrefresh(debug2_win);

	/* only restore if saved because text memory might be less than 4096 bytes */
	if (save_char != '\0')
		text[4095] = save_char;
}


/**********************************************************************/
static void display_debug3(char *text)
{
static	char	*win_title = "Debug3 Window";
	char	save_char = '\0';

	if (strlen(text) > 4095) {
		save_char = text[4095];
		text[4095] = '\0';
	}

	werase(debug3_win);
	mvwprintw(debug3_win, 4, 2, text);
	box(debug3_win, '*', '*');
	mvwprintw(debug3_win, 0, (22 - strlen(win_title)/2), " %s ", win_title);
	wrefresh(debug3_win);

	/* only restore if saved because text memory might be less than 4096 bytes */
	if (save_char != '\0')
		text[4095] = save_char;
}


/**********************************************************************/
static void display_debug3_top(char *text)
{
static	char	*win_title = "Debug3 Window";
	char	save_char = '\0';

	if (strlen(text) > 4095) {
		save_char = text[4095];
		text[4095] = '\0';
	}

#if 0
	werase(debug3_win);
#endif
	mvwprintw(debug3_win, 2, 2, text);
	box(debug3_win, '*', '*');
	mvwprintw(debug3_win, 0, (22 - strlen(win_title)/2), " %s ", win_title);
	wrefresh(debug3_win);

	/* only restore if saved because text memory might be less than 4096 bytes */
	if (save_char != '\0')
		text[4095] = save_char;
}


/**********************************************************************/
static int display_init(char *term)
{
	int	debug1_size_x,
		debug1_size_y;

	/* in this version, a NULL term implies not to display anything */
	if (term == NULL) {
		return(0);
	}

	if (term == NULL) {
		/* If no terminal string is given, then use the main screen */
		/*   (unless we catch this case above)                      */
		initscr();
	} else {
		/* TODO: figure out how to automatically suspend the shell now controlling the term */
		FILE	*term_fp;

		term_fp = fopen(term, "a+");
		if (term_fp != NULL) {
			scr = newterm(NULL, term_fp, term_fp);
#if 0
			def_prog_mode();
#endif
			set_term(scr);
			clear_screen();
		} else {
			fprintf(stderr, "Could not open terminal '%s' for curses display.\n", term);
			return(0);
		}
	}

	crmode();
	noecho();
	refresh();

	debug1_size_x = COLS - 4;
	debug1_size_y = (LINES - 4)/4;

	debug1_win = newwin(debug1_size_y, debug1_size_x, 0, 0);
	debug2_win = newwin(debug1_size_y, debug1_size_x, debug1_size_y+1, 0);
	debug3_win = newwin(debug1_size_y*2+2, debug1_size_x, (debug1_size_y*2)+1, 0);

	/* render the windows so the boxes will show up */
	display_debug1("");
	display_debug2("");
	display_debug3("");

	return (1);
}


/**********************************************************************/
static void display_term()
{
	if (scr != NULL) {
		mvcur(0, COLS-1, 23, 0);
		endwin();
	}
}


#undef TRUE	/* defined by curses, but Performer uses it too */
#undef FALSE	/* defined by curses, but Performer uses it too */
#undef clear	/* defined by curses, but Performer uses it too */
#undef move	/* defined by curses, but Performer uses it too */


/*****************************************************************************/
/*** ===================================================================== ***/
/*****************************************************************************/

/* Performer Node, Path and Scene informational routines */

#include <Performer/pf/pfDCS.h>
#include <Performer/pf/pfSequence.h>
#include <Performer/pf/pfSwitch.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLOD.h>
#include <Performer/pr/pfTexture.h>
#if 0 /* I may want to use these in the near future */
#  include <Performer/pr/pfMaterial.h>
#  include <Performer/pr/pfLinMath.h>	/* for pfMatrix functions */
#endif

#if 0
/* NOTE: these text color modifiers are currently unused because */
/*       they don't work in curses.                              */
#define NORM_TEXT       "^[[0m"
#define BOLD_TEXT       "^[[1m"
#define RED_TEXT        "^[[31m"
#endif


/************************************************************************/
/*  modify_path(): takes a pfPath and movement instructions.  The       */
/*    path is modified based on the instructions, and the Node at the   */
/*    end of the path is returned.                                      */
/*                                                                      */
/*  Movement operations are based via two integer operations.		*/
/*    Five "movement" operations are possible:                          */
/*      0, <n> -- nop (ie. no movement), second argument is ignored     */
/*      1, <n> -- move right by <n> nodes                               */
/*      2, <n> -- move left by <n> nodes                                */
/*      3, <n> -- move down one level, to child number <n>              */
/*      4, <n> -- move up one level, second argument is ignored         */
static pfNode *modify_path(pfPath *path, int movement, int move_index)
{
	pfNode		*last_node;
	pfNode		*root_node;
	pfNode		*parent_node;
	int		path_len;
	int		index;

	path_len = (int)path->getNum();
	root_node = (pfNode *)path->get(0);
	last_node = (pfNode *)path->get(path_len-1);
	if (last_node != root_node)
		parent_node = (pfNode *)path->get(path_len-2);
	else	parent_node = NULL;

	switch (movement) {

	case 0: /* nop -- just return the last node in the path */
		break;

	case 1: /* right -- select next child */
		if (parent_node != NULL) {
			index = ((pfGroup *)parent_node)->searchChild(last_node);
			if (index+move_index < ((pfGroup *)parent_node)->getNumChildren()) {
				path->set(path_len-1, ((pfGroup *)parent_node)->getChild(index+move_index));
			}
		}
		break;

	case 2: /* left -- select previous child */
		if (parent_node != NULL) {
			index = ((pfGroup *)parent_node)->searchChild(last_node);
			if (index-move_index >= 0) {
				path->set(path_len-1, ((pfGroup *)parent_node)->getChild(index-move_index));
			}
		}
		break;

	case 3: /* down -- move down to selected child node */
		if (last_node->isOfType(pfGroup::getClassType())) {
			/* make sure at least one child exists, and add first child */
			if (((pfGroup *)last_node)->getNumChildren() > move_index)
				path->add(((pfGroup *)last_node)->getChild(move_index));
		}
		break;

	case 4: /* up -- move up in path (i.e. remove last node) */
		if (last_node != root_node) {
			path->removeIndex(path_len-1);
		}
		break;
	}

	return ((pfNode *)path->get(path->getNum()-1));
}


/**********************************************************************/
static char *pfPathString(pfPath *path, char *text, int max_len)
{
	int		count;
	int		path_len;
	int		index = -1;
	pfNode		*node;
	pfNode		*last_node = NULL;

	path_len = path->getNum();
	snprintf(text, max_len, "Path: length %d\n", path_len);

	for (count = 0; count < path_len; count++) {
		node = (pfNode *)path->get(count);
		if (last_node != NULL) {
			index = ((pfGroup *)last_node)->searchChild(node);
		}
		snprintf(text, max_len, "%s\tNode %d (0x%p): a %s named '%s', %3dth child of above",
			text, count, node, node->getTypeName(), node->getName(), index);
#if 0 /* TODO: need to figure out args to pfLOD::evaluate(...) in order to report     */
	/* which LOD value is active, or maybe this isn't possible in the app process.*/
		if (node->isOfType(pfLOD::getClassType())) {
			snprintf(text, max_len, "%s\t\tLOD->evaluate = %5.3f\n", text,
				((pfLOD *)node)->evaluate());
		}
#endif
		if (node->isOfType(pfGroup::getClassType()))
			snprintf(text, max_len, "%s -- has %d children\n", text, ((pfGroup *)node)->getNumChildren());
		else	snprintf(text, max_len, "%s\n", text);
		last_node = node;
	}

	return (text);	/* duplicate call by ref return */
}


/**********************************************************************/
/* create a string with the types/names of the nearby siblings of the */
/*   given node -- useful for debugging scenegraphs.                  */
static char *pfNodeSiblingsString(pfNode *node, int parent_num, char *text, int max_len)
{
	pfGroup		*parent;
	pfNode		*child;
	int		num_children;
	int		node_index;
	int		count;
	char		sibstring[128];

	if (node->getNumParents() <= parent_num) {
		snprintf(text, max_len, "   *** Node does not have a parent %d ***\n", parent_num);
		return (text);
	}

	text[0] = '\0';

	parent = node->getParent(parent_num);
	num_children = parent->getNumChildren();
	node_index = parent->searchChild(node);

	for (count = node_index - 2; count <= node_index + 2; count++) {
		if (count == node_index) {
#if 0 /* this doesn't work */
			snprintf(text, max_len, "%s " BOLD_TEXT "[*]" NORM_TEXT, text);
#else
			snprintf(text, max_len, "%s " "[*]" , text);
#endif
		} else if (count >= 0 && count < num_children) {
			child = parent->getChild(count);
			sprintf(sibstring, "%.10s(%s)", child->getTypeName(), child->getName());
			if (sibstring[22] != '\0') {
				sibstring[22] = ')';	/* limit length of string */
				sibstring[23] = '\0';
			}
			snprintf(text, max_len, "%s [%-22.22s]", text, sibstring);
		} else {
			/* count is outside range of children */
			snprintf(text, max_len, "%s [    --    (          )]", text);
		}
	}

	/* add a carriage return */
	snprintf(text, max_len, "%s\n", text);

	return (text);	/* duplicate call by ref return */
}


/**********************************************************************/
/* Display information about the given pfGeoState. */
/* Currenly only displaying the name of the active texture image. */
static char *pfGeoStateString(pfGeoState *geostate, char *text, int max_len)
{
	pfTexture	*tex;

	if (geostate == NULL) {
#if 1
		snprintf(text, max_len, "= null");
#endif
		return text;
	}

	tex = (pfTexture *)(geostate->getAttr(PFSTATE_TEXTURE));

	snprintf(text, max_len, "0x%p = [tex = 0x%p (%s)]", text,
		tex, (tex == NULL ? "--" : (char *)((pfTexture *)tex)->getName()));

	return text;
}


/**********************************************************************/
/* NOTE: here are the current values of the primative types: */
/*	0 - PFGS_POINTS           */
/*	1 - PFGS_LINES            */
/*	2 - PFGS_LINESTRIPS       */
/*	3 - PFGS_TRIS             */
/*	4 - PFGS_QUADS            */
/*	5 - PFGS_TRISTRIPS        */
/*	6 - PFGS_FLAT_LINESTRIPS  */
/*	7 - PFGS_FLAT_TRISTRIPS   */
/*	8 - PFGS_POLYS            */
/*	9 - PFGS_TRIFANS          */
/*	10 - PFGS_FLAT_TRIFANS    */
static char *pfNodeString(pfNode *node, void *match, int prim_num, int skip_children, char *text, int max_len)
{
static	char		*prim_name[] = {
				"points", "lines", "linestrips", "tris", "quads",
				"tristrips", "flat linestrips", "flat tristrips",
				"polys", "trifans", "flat trifans",
				"unknown", "unknown 2", "unknown 3", "unknown 4",
				"unknown 5", "unknown 6", "unknown 7", "unknown 8"
			};
static	char		*bind_name[] = {
				"off", "overall", "per_prim", "per_vert"
			};
static	char		tmpbuf[1024];
	int		count, count2, count3;
	int		num_children, num_items;
	int		temp_int;
	char		*slot_name;
	pfGroup		*group;
	pfNode		*child;
	pfGeoSet	*geoset;
	pfObject	*node_as_object;
	pfNodeTravFuncType	prefunc;
	pfNodeTravFuncType	postfunc;
	pfMatrix	matrix;
	float		x, y, z, w;

	snprintf(text, max_len, "Info for node 0x%p:", node);

	if (node == NULL)
		return (text);

	/************************/
	/** Node Type and Name **/
	snprintf(text, max_len, "%s  type = '%s',  name = '%s'\n", text,
		node->getTypeName(), node->getName());

	/**************************/
	/** Node Traversal Masks **/
	node->getTravFuncs(PFTRAV_ISECT, &prefunc, &postfunc);
	snprintf(text, max_len, "%s\tIsect traversal mask = %08X mode = %d data = 0x%p precb = 0x%p postcb = 0x%p\n",
		text, node->getTravMask(PFTRAV_ISECT), node->getTravData(PFTRAV_ISECT), /* node->getTravMode(PFTRAV_ISECT, 1) */(void *)NULL,
		&prefunc, &postfunc);

/****** TODO: add data, precb & postcb for the following traversals ******/

	snprintf(text, max_len, "%s\tDraw  traversal mask = %X\n", text, node->getTravMask(PFTRAV_DRAW));
	snprintf(text, max_len, "%s\tCull  traversal mask = %X\n", text, node->getTravMask(PFTRAV_CULL));
	snprintf(text, max_len, "%s\tApp   traversal mask = %X\n", text, node->getTravMask(PFTRAV_APP));

	/********************/
	/** Node User Data **/
	node_as_object = ((pfObject *)node);
	temp_int = node_as_object->getNumUserData();
	snprintf(text, max_len, "%s\tUser Data Slots = %d\n", text, temp_int);
	for (count = 0; count < temp_int; count ++)
		snprintf(text, max_len, "%s\t\tUser Data Slots %d = 0x%p\n", text, count, node_as_object->getUserData(count));

	/**************************/
	/** Node Named User Data **/
	temp_int = node_as_object->getNumNamedUserDataSlots();
	snprintf(text, max_len, "%s\tNamed User Data Slots = %d\n", text, temp_int);
	for (count = 0; count < temp_int; count ++) {
		slot_name = (char *)node_as_object->getUserDataSlotName(count);
		snprintf(text, max_len, "%s\t\tNamed User Data Slots %d = '%s' for slot %d\n", text, count,
			slot_name, (int)node_as_object->getNamedUserDataSlot(slot_name));
	}

	/***************************/
	/** Node Matrix transform **/
	if (node->isOfType(pfSCS::getClassType())) {
		/* print matrix information for pfSCS & pfDCS (a subclass) nodes */

		((pfSCS *)node)->getMat(matrix);
		for (count = 0; count < 4; count++) {
			matrix.getRow(count, &x, &y, &z, &w);
			snprintf(text, max_len, "%s\t%s| %11.4f %11.4f %11.4f %11.4f |\n", text,
				(count == 0 ? "matrix = " : "         "), x, y, z, w);
		}
	}

	/************************/
	/** Node Children info **/
	if (node->isOfType(pfGroup::getClassType())) {

		/***********************************/
		/* print Group related information */

		group = (pfGroup *)node;

		/* first print specific info based on the type of group */

		if (node->isOfType(pfLOD::getClassType())) {
			/* print LOD information */

/*********************************************************/
/* Lots of experiments in showing the current LOD choice */
#if 0 /* { */
			pfMatrix	*offset_mat = NULL;

#include <pfcave.h>	/* not sure if we need this (currently only needed for pfCAVEMasterChan()) */

			pfCAVEMasterChan()->getOffsetViewMat(matrix);
	for (count = 0; count < 4; count++) {
		matrix.getRow(count, &x, &y, &z, &w);
		snprintf(text, max_len, "%s\t%s| %11.4f %11.4f %11.4f %11.4f |\n", text,
			(count == 0 ? "off mat= " : "         "), x, y, z, w);
	}


#if 0
	matrix.preMult(*(pfMatrix *)((pfDCS *)isect_node)->getMatPtr()); /* bad */
	matrix.preMult((pfMatrix)((pfDCS *)isect_node)->getMatPtr());	/* won't compile */
#else
	((pfDCS *)world_dcs)->getMat(*offset_mat);
	matrix.preMult(*offset_mat);
#endif
	for (count = 0; count < 4; count++) {
		matrix.getRow(count, &x, &y, &z, &w);
		snprintf(text, max_len, "%s\t%s| %11.4f %11.4f %11.4f %11.4f |\n", text,
			(count == 0 ? "premultmat= " : "         "), x, y, z, w);
	}
			snprintf(text, max_len, "%s\tLOD premult choice = %d\n", text,
				(int)((float)((pfLOD *)group)->evaluate(pfCAVEMasterChan(), &matrix)));


	((pfDCS *)world_dcs)->getMat(*offset_mat);
	matrix.postMult(*offset_mat);
	for (count = 0; count < 4; count++) {
		matrix.getRow(count, &x, &y, &z, &w);
		snprintf(text, max_len, "%s\t%s| %11.4f %11.4f %11.4f %11.4f |\n", text,
			(count == 0 ? "postmultmat= " : "         "), x, y, z, w);
	}
			snprintf(text, max_len, "%s\tLOD premult choice = %d\n", text,
				(int)((float)((pfLOD *)group)->evaluate(pfCAVEMasterChan(), &matrix)));


	((pfDCS *)world_dcs)->getMat(*offset_mat);
	matrix.postMult(*offset_mat);
	if ((((pfNode *)group)->getParent(0))->isOfType(pfSCS::getClassType())) {
	((pfDCS *)((pfNode *)group)->getParent(0))->getMat(*offset_mat);
	matrix.preMult(*offset_mat);
	for (count = 0; count < 4; count++) {
		matrix.getRow(count, &x, &y, &z, &w);
		snprintf(text, max_len, "%s\t%s| %11.4f %11.4f %11.4f %11.4f |\n", text,
			(count == 0 ? "parent pre= " : "         "), x, y, z, w);
	}
			snprintf(text, max_len, "%s\tLOD parent pre mult choice = %d\n", text,
				(int)((float)((pfLOD *)group)->evaluate(pfCAVEMasterChan(), &matrix)));
	}


			pfCAVEMasterChan()->getViewMat(*offset_mat);
	for (count = 0; count < 4; count++) {
		offset_mat->getRow(count, &x, &y, &z, &w);
		snprintf(text, max_len, "%s\t%s| %11.4f %11.4f %11.4f %11.4f |\n", text,
			(count == 0 ? "viewmat= " : "         "), x, y, z, w);
	}
			snprintf(text, max_len, "%s\tLOD (viewmat offset mat) choice = %d\n", text,
				(int)((float)((pfLOD *)group)->evaluate(pfCAVEMasterChan(), offset_mat)));


#endif /* } LOD experiments */
/*********************************************************/



#if 0 /* currently not found to be useful */
			snprintf(text, max_len, "%s\tLOD (state index) choice = %d\n", text,
				(int)((pfLOD *)group)->getLODStateIndex());
#endif
			num_items = ((pfLOD *)group)->getNumRanges();
			snprintf(text, max_len, "%s\tNum Ranges = %d\n\t     Range 0 =%12.4f\n", text, num_items, ((pfLOD *)group)->getRange(0));

		} else if (node->isOfType(pfSequence::getClassType())) {
			/* print Sequence information */

			snprintf(text, max_len, "%s\tFrame = %d\n", text, ((pfSequence *)group)->getFrame(&temp_int));

		} else if (node->isOfType(pfSwitch::getClassType())) {
			/* print Switch information */

			temp_int = (int)(((pfSwitch *)group)->getVal());
			snprintf(text, max_len, "%s\tSwitch value = %s (%d)\n", text,
				(temp_int == PFSWITCH_ON ? "PFSWITCH_ON" : (temp_int == PFSWITCH_OFF ? "PFSWITCH_OFF" : "unknown")), temp_int);
		}

		/* now print info about each child */
		num_children = group->getNumChildren();
		snprintf(text, max_len, "%s\t%d children:\n", text, num_children);
		for (count = skip_children; count < num_children; count++) {
			child = group->getChild(count);
			snprintf(text, max_len, "%s\t\t%3d: (0x%p)", text, count, child);

			if (node->isOfType(pfLOD::getClassType())) {
				/* print LOD information */

				snprintf(text, max_len, "%s%13.4f :", text, ((pfLOD *)group)->getRange(count+1));
			} else if (node->isOfType(pfSequence::getClassType())) {
				snprintf(text, max_len, "%s%10.4lf :", text, ((pfSequence *)group)->getTime(count));
			}

			snprintf(text, max_len, "%s a %s named '%s'\n", text,
				child->getTypeName(), child->getName());
		}
	} else if (node->isOfType(pfGeode::getClassType())) {

		/***********************************/
		/* print Geode related information */

		num_items = ((pfGeode *)node)->getNumGSets();
		snprintf(text, max_len, "%s\t%d pfGeoSet(s)\n", text, num_items);
		for (count = skip_children; count < num_items; count++) {
			geoset = ((pfGeode *)node)->getGSet(count);
			if (geoset == (pfGeoSet *)match) {
				/* NOTE: we don't currenlty use the match feature, and I'm not  */
				/*   sure what my original intention was.  So, this part of the */
				/*   code is not fully implemented.  Basically need to copy the */
				/*   code in the else clause if it's decided to implement this. */
				snprintf(text, max_len, "%s\t%2d of --> GeoSet %d (0x%p): has %d %s, imask = %04x, GSt %s\n",
					text, prim_num, count,
					geoset, geoset->getNumPrims(), prim_name[geoset->getPrimType()], geoset->getIsectMask(),
					pfGeoStateString(geoset->getGState(), tmpbuf, 1023));
#if 0 /* [6/22/2000] not sure when we get here -- and doesn't work (see below) */
/* this should work now, so need to make it */
/*   match the code in the else clause.     */
				/* add info about the geoset verticies */
				pfVec3	*vertlist;
				geoset->getAttrLists(PFGS_COORD3, (void **)(&vertlist), NULL);
				snprintf(text, max_len, "%s\tfirst vertice at (%f %f %f)\n", text, vertlist[0].vec[0], vertlist[0].vec[1], vertlist[0].vec[2]);
#endif
			} else {
				int		color_bind, vert_bind, norm_bind, tex_bind;
				pfVec4		*colorlist;
				pfVec3		*vertlist;
				pfVec3		*normlist;
				pfVec2		*texlist;
				unsigned short	*colorindex;
				unsigned short	*vertindex;
				unsigned short	*normindex;
				unsigned short	*texindex;
				int		*lengths;
				int		last_vert;


#if 0 /* get values for debugging */
snprintf(text, max_len, "%s\n\tPFGS_POINTS = %d, PFGS_LINES = %d, PFGS_LINESTRIPS = %d, PFGS_FLAT_LINESTRIPS = %d, PFGS_TRIS = %d, PFGS_QUADS = %d\n PFGS_TRISTRIPS = %d, PFGS_FLAT_TRISTRIPS = %d, PFGS_TRIFANS = %d, PFGS_FLAT_TRIFANS = %d, PFGS_POLYS = %d\n",
text, PFGS_POINTS, PFGS_LINES, PFGS_LINESTRIPS, PFGS_FLAT_LINESTRIPS, PFGS_TRIS, PFGS_QUADS, PFGS_TRISTRIPS, PFGS_FLAT_TRISTRIPS, PFGS_TRIFANS, PFGS_FLAT_TRIFANS, PFGS_POLYS);
#endif
				/* first print the basic info about the node */
				snprintf(text, max_len, "%s\t\tGSet %d (0x%p): has %d %s, imask = %04x,  GSt %s\n",
					text, count,
					geoset, geoset->getNumPrims(), prim_name[geoset->getPrimType()], geoset->getIsectMask(),
					pfGeoStateString(geoset->getGState(), tmpbuf, 1023));

				/* print how each attribute is bound */
				color_bind = geoset->getAttrBind(PFGS_COLOR4);
				vert_bind = geoset->getAttrBind(PFGS_COORD3);
				norm_bind = geoset->getAttrBind(PFGS_NORMAL3);
				tex_bind = geoset->getAttrBind(PFGS_TEXCOORD2);
				snprintf(text, max_len, "%s\t\t\tBindings: color-%s, vertices-%s, normals-%s, texcoords-%s\n", text, bind_name[color_bind], bind_name[vert_bind], bind_name[norm_bind], bind_name[tex_bind]);

				/* print the overall bindings */
				/* NOTE: this doesn't make sense for anything but color, but this IS a debugger */
		geoset->getAttrLists(PFGS_COLOR4, (void **)(&colorlist), NULL);
				if (color_bind == PFGS_OVERALL)
					snprintf(text, max_len, "%s\t\t\toverall color: (%5.3f %5.3f %5.3f %5.3f)\n", text, colorlist[0].vec[0], colorlist[0].vec[1], colorlist[0].vec[2], colorlist[0].vec[3]);
				if (vert_bind == PFGS_OVERALL)
					snprintf(text, max_len, "%s\t\t\toverall vertex: (%5.3f %5.3f %5.3f)\n", text, 0.0, 0.0, 0.0);
				if (norm_bind == PFGS_OVERALL)
					snprintf(text, max_len, "%s\t\t\toverall normal: (%5.3f %5.3f %5.3f)\n", text, 0.0, 0.0, 0.0);
				if (tex_bind == PFGS_OVERALL)
					snprintf(text, max_len, "%s\t\t\toverall tex coord: (%5.3f %5.3f)\n", text, 0.0, 0.0);

				/* now get list values for attributes with lists */
				if (color_bind == PFGS_PER_PRIM || color_bind == PFGS_PER_VERTEX)
					geoset->getAttrLists(PFGS_COLOR4, (void **)(&colorlist), &colorindex);
				if (vert_bind == PFGS_PER_PRIM || vert_bind == PFGS_PER_VERTEX)
					geoset->getAttrLists(PFGS_COORD3, (void **)(&vertlist), &vertindex);
				if (norm_bind == PFGS_PER_PRIM || norm_bind == PFGS_PER_VERTEX)
					geoset->getAttrLists(PFGS_NORMAL3, (void **)(&normlist), &normindex);
				if (tex_bind == PFGS_PER_PRIM || tex_bind == PFGS_PER_VERTEX)
					geoset->getAttrLists(PFGS_TEXCOORD2, (void **)(&texlist), &texindex);
				lengths = geoset->getPrimLengths();

				switch (geoset->getPrimType()) {

				default:
					/* just print the first vertex by default */
					snprintf(text, max_len, "%s\t\t\tfirst vertex at (%f %f %f)\n", text, vertlist[0].vec[0], vertlist[0].vec[1], vertlist[0].vec[2]);
					break;

				case PFGS_QUADS:	/* (4)  see if this works too! */
				case PFGS_POLYS:	/* (8)  see if this works too! */
				case PFGS_TRISTRIPS:	/* (5) */
					last_vert = 0;
					for (count2 = 0; count2 < geoset->getNumPrims(); count2++) {
						/* first print the per-primative information */
						snprintf(text, max_len, "%s\t\t\tStrip %d: %d vertices", text, count2, lengths[count2]);
						if (color_bind == PFGS_PER_PRIM)
							snprintf(text, max_len, "%s  Color: (%5.3f %5.3f %5.3f %5.3f)", text, colorlist[count2].vec[0], colorlist[count2].vec[1], colorlist[count2].vec[2], colorlist[count2].vec[3]);
						if (vert_bind == PFGS_PER_PRIM)
							snprintf(text, max_len, "%s  Vertex: (%5.3f %5.3f %5.3f)", text, vertlist[count2].vec[0], vertlist[count2].vec[1], vertlist[count2].vec[2]);
						if (norm_bind == PFGS_PER_PRIM)
							snprintf(text, max_len, "%s  Normal: (%5.3f %5.3f %5.3f)", text, normlist[count2].vec[0], normlist[count2].vec[1], normlist[count2].vec[2]);
						if (tex_bind == PFGS_PER_PRIM)
							snprintf(text, max_len, "%s  Tex Coord: (%5.3f %5.3f)", text, texlist[count2].vec[0], texlist[count2].vec[1]);

						/* now print the per-vertex information */
						if (vertindex != NULL) {
							/* for pfGeoSet primatives using non-indexed vertices (mixing not allowed) */
							for (count3 = last_vert; count3 < lengths[count2]+last_vert; count3++) {
								snprintf(text, max_len, "%s\n\t\t\t\tvertex %2d: (%7.3f %7.3f %7.3f)", text, count3, vertlist[count3].vec[0], vertlist[count3].vec[1], vertlist[count3].vec[2]);
								if (color_bind == PFGS_PER_VERTEX)
									snprintf(text, max_len, "%s  Color: (%5.3f %5.3f %5.3f %5.3f)", text, colorlist[count3].vec[0], colorlist[count3].vec[1], colorlist[count3].vec[2], colorlist[count3].vec[3]);
								if (norm_bind == PFGS_PER_VERTEX)
									snprintf(text, max_len, "%s  Normal: (%5.3f %5.3f %5.3f)", text, normlist[count3].vec[0], normlist[count3].vec[1], normlist[count3].vec[2]);
								if (tex_bind == PFGS_PER_VERTEX)
									snprintf(text, max_len, "%s  Tex: (%5.3f %5.3f)", text, texlist[count3].vec[0], texlist[count3].vec[1]);
							}
							snprintf(text, max_len, "%s\n", text);
							last_vert = count3;
						} else {
							/* for pfGeoSet primatives using indexed vertices */
							snprintf(text, max_len, "%s(indexed primative - debug print not yet implemented)\n", text);
						}
					}
					break;
				}
			}
#if 0
	/* some old debugging code, I'm not sure what I used it for */
		geoset->getGState()->print(PFTRAV_SELF | PFTRAV_DESCEND, PFPRINT_VB_DEBUG, "gstate: ", NULL);
				abort();
		((pfMemory *)last_node)->print(PFTRAV_SELF | PFTRAV_DESCEND, PFPRINT_VB_DEBUG, "yo: ", NULL);
		geoset->print(PFTRAV_SELF | PFTRAV_DESCEND, PFPRINT_VB_DEBUG, "yoyo: ", NULL);
		((pfMemory *)tex)->print(PFTRAV_SELF, PFPRINT_VB_DEBUG, "tex: ", NULL);
#endif
		}

	}

	return (text);	/* duplicate call by ref return */
	}


	/**********************************************************************/
	static char *pfNodeAncestryString(pfNode *node, pfNode *adam, char *text, int max_len)
{
	snprintf(text, max_len, "Ancestry of node 0x%p\n", node);

	do {
		snprintf(text, max_len, "%s\tNode is 0x%p, type = '%s', name = '%s'\n",
			text, node, node->getTypeName(), node->getName());
		if (node->getNumParents() == 0) {
			snprintf(text, max_len, "%s\tRoot of the tree\n", text);
			node = NULL;
		} else {
			node = node->getParent(0);
		}
	} while (node != NULL && node != adam);

	if (node != NULL) {
		/* print adam */
		snprintf(text, max_len, "%s\tNode is 0x%p, type = '%s', name = '%s'\n",
			text, node, node->getTypeName(), node->getName());
	}

	return (text);	/* duplicate call by ref return */
}


/**********************************************************************/
static void pfNodeGraphPrint(FILE *fp, char *indent, pfNode *node)
{
	pfGroup	*group;
	char	*new_indent;
	int	num_children;
	int	count;

	fprintf(fp, "%s%s ('%s')\n", indent, node->getTypeName(), node->getName());

	if (node->isOfType(pfGroup::getClassType())) {
		new_indent = (char *)malloc(strlen(indent + 3));
		sprintf(new_indent, "%s  ", indent);

		group = (pfGroup *)node;
		num_children = group->getNumChildren();
		for (count = 0; count < num_children; count++) {
			pfNodeGraphPrint(fp, new_indent, group->getChild(count));
		}

#if 0	/* causes core dump for some (unknown) reason */
		free(new_indent);
#endif
	}
}



/*****************************************************************************/
/*** ===================================================================== ***/
/*****************************************************************************/


/*******************************************************************/
/* These are the publically available routines to initialize and   */
/*   close the display and manipulates and render the scenegraph.  */
/* NOTE that this is the only section that requires any references */
/*   to the VR library -- and even that could be reduced fairly  */
/*   trivially by passing an array of "button" values (ie. the     */
/*   debouncing could still be done here.)                         */

#include <freevr.h>	/* needed for button functions and definitions */


/************************************************************************/
int pfdebug_initialize_display(char *tty)
{
	return(display_init(tty));
}


/************************************************************************/
void pfdebug_terminate_display()
{
	display_term();
}

/* a simple macro for debouncing button inputs */
#define	DEBOUNCE(key, prev) ((vrGet2switchValue((key))) ? (!(prev) ? (prev) = 1 : 0) : ((prev) = 0))



/************************************************************************/
void pfdebug_display_sg(pfScene *scene)
{
static	char	buffer[32768];		/* This is static just to prevent constant reallocation */
static	char	node_info[1024];	/* ditto */
static	int	skip_children = 0;
	int	prev_skip_children;
static	pfPath	*scene_path = NULL;	/* A path through the scenegraph */
static	pfNode	*cut_node = NULL;	/* storage for a node that has been cut (for pasting) */
static	pfNode	*empty_node = NULL;	/* pointer to a temporary replacement node of the disabled node */
static	pfNode	*last_node = NULL;	/* the last node in the path */
	pfNode	*prev_last_node = NULL;	/* the previous last node */

	/* Keys to use for various functions (and previous values for debouncing) */
static	int	key_movedown = 13,		prev_movedown = 0;
static	int	key_moveup = 14,		prev_moveup = 0;
static	int	key_moveleft = 12,		prev_moveleft = 0;
static	int	key_moveright = 15,		prev_moveright = 0;
static	int	key_skiptenmore = 23,		prev_skiptenmore = 0;
static	int	key_skiptenless = 22,		prev_skiptenless = 0;
static	int	key_skiponemore = 19,		prev_skiponemore = 0;
static	int	key_skiponeless = 20,		prev_skiponeless = 0;
static	int	key_skipnone = 16,		prev_skipnone = 0;
static	int	key_skiptoend = 17,		prev_skiptoend = 0;
static	int	key_printgraph = 18,		prev_printgraph = 0;
static	int	key_printnode = 7,		prev_printnode = 0;
static	int	key_hidenode = 11,		prev_hidenode = 0;
static	int	key_cutnode = 24,		prev_cutnode = 0;
static	int	key_pastenode = 25,		prev_pastenode = 0;
static	int	key_shiftleft = 8,		prev_shiftleft = 0;
static	int	key_shiftright = 9,		prev_shiftright = 0;

	/*******************************************/
	/* initialize some of the static variables */
	if (scene_path == NULL) {
		scene_path = new pfPath;
		scene_path->add(scene);
	}
	if (last_node == NULL)
		last_node = modify_path(scene_path, 0, 0);

	if (empty_node == NULL) {
		empty_node = (pfNode *)new pfGroup;
		empty_node->setName("unused empty node");
	}

	prev_last_node = last_node;
	prev_skip_children = skip_children;

	/******************************************************/
	/*** Keys to control the PF scene graph curses view ***/

	/*** move down ***/
	if (DEBOUNCE(key_movedown, prev_movedown)) {
		last_node = modify_path(scene_path, 3, skip_children);
		skip_children = 0;

		display_debug2(pfPathString(scene_path, buffer, 32768));
	}

	/*** move up ***/
	if (DEBOUNCE(key_moveup, prev_moveup)) {
		last_node = modify_path(scene_path, 4, 0);
		skip_children = 0;

		display_debug2(pfPathString(scene_path, buffer, 32768));
	}

	/*** move to the right ***/
	if (DEBOUNCE(key_moveright, prev_moveright)) {
		last_node = modify_path(scene_path, 1, 1);
		skip_children = 0;

		display_debug2(pfPathString(scene_path, buffer, 32768));
	}

	/*** move to the left ***/
	if (DEBOUNCE(key_moveleft, prev_moveleft)) {
		last_node = modify_path(scene_path, 2, 1);
		skip_children = 0;

		display_debug2(pfPathString(scene_path, buffer, 32768));
	}

	/*** skip 10 more children ***/
	if (DEBOUNCE(key_skiptenmore, prev_skiptenmore)) {
		skip_children += 10;
	}

	/*** skip 1 more child ***/
	if (DEBOUNCE(key_skiponemore, prev_skiponemore)) {
		skip_children++;
	}

	/*** go back 10 children ***/
	if (DEBOUNCE(key_skiptenless, prev_skiptenless)) {
		skip_children -= 10;
		if (skip_children < 0)
			skip_children = 0;
	}

	/*** go back 1 child ***/
	if (DEBOUNCE(key_skiponeless, prev_skiponeless)) {
		if (skip_children > 0)
			skip_children--;
	}

	/*** skip 0 children ***/
	if (DEBOUNCE(key_skipnone, prev_skipnone)) {
		skip_children = 0;
	}

	/*** skip to bottom child (or close to bottom) ***/
	if (DEBOUNCE(key_skiptoend, prev_skiptoend)) {
		if (last_node->isOfType(pfGroup::getClassType()))
			skip_children = ((pfGroup *)last_node)->getNumChildren();
		else if (last_node->isOfType(pfGeode::getClassType()))
			skip_children = ((pfGeode *)last_node)->getNumGSets();
		else    skip_children = 0;

		switch (skip_children) {
		case 0: /* there are no children */
			break;

		case 1: /* show all the children */
		case 2:
		case 3:
			skip_children = 0;


		default:/* show the last few children */
			skip_children -= 4;
		}
	}

	/*** print a simple representation of the graph from this node, below ***/
	if (DEBOUNCE(key_printgraph, prev_printgraph)) {
		fprintf(stdout, "/------------------------------\n");
		pfNodeGraphPrint(stdout, "| ", last_node);
		fprintf(stdout, "\\------------------------------\n");
	}

	/*** print the node-detail information to stdout ***/
	if (DEBOUNCE(key_printnode, prev_printnode)) {
		fprintf(stdout, "/----------------------------------------------------------------------\n");
		fprintf(stdout, pfNodeString(last_node, NULL, 0, skip_children, buffer, 32768));
		fprintf(stdout, "\\----------------------------------------------------------------------\n");
	}


	/*****************************************/
	/*** Keys to modify the PF scene graph ***/

	/*** toggle the inclusion of the current node in the scenegraph ***/
	if (DEBOUNCE(key_hidenode, prev_hidenode)) {
	static	pfNode	*hidden_node = NULL;		/* storage for a temporarily disabled node */
	static	pfGroup	*hidden_parent_node = NULL;	/* storage for parent of disabled node */

		if (hidden_node == NULL) {
			/* temporarily remove the node */
			int	path_len;
			path_len = (int)scene_path->getNum();
			if (path_len < 2) {
				/* we can't change remove the only node */
			} else {
				hidden_parent_node = (pfGroup *)scene_path->get(path_len-2);
				hidden_node = (pfNode *)scene_path->get(path_len-1);

				hidden_parent_node->replaceChild(last_node, empty_node);

				sprintf(node_info, "Hidden node 0x%x '%s' is of type %s", hidden_node, hidden_node->getName(), hidden_node->getTypeName());
				display_debug1_line(2, node_info);
			}
		} else {
			/* restore the node */
			hidden_parent_node->replaceChild(empty_node, hidden_node);
			hidden_node = NULL;
			display_debug1_line(2, "");
		}
	}

	/*** cut a node out of the scenegraph ***/
	if (DEBOUNCE(key_cutnode, prev_cutnode)) {
		int	path_len;
		pfGroup	*parent_node;

		path_len = (int)scene_path->getNum();
		if (path_len < 2) {
			/* we can't cut the scene node */
		} else {
			parent_node = (pfGroup *)scene_path->get(path_len-2);
			cut_node = last_node;
			parent_node->removeChild(cut_node);
			last_node = modify_path(scene_path, 4, 0);
			display_debug2(pfPathString(scene_path, buffer, 32768));

			sprintf(node_info, "Cut node 0x%x '%s' is of type %s", cut_node, cut_node->getName(), cut_node->getTypeName());
			display_debug1_line(4, node_info);
		}
	}

	/*** paste the last cut-node into the scenegraph ***/
	/*   NOTE: this only inserts before the current child, so to add  */
	/*         at the end of a list of children a right shift is nece.*/
	if (DEBOUNCE(key_pastenode, prev_pastenode)) {
		int	path_len;
		int	last_node_index;
		pfGroup	*parent_node;

		path_len = (int)scene_path->getNum();
		if (path_len < 2) {
			/* we can't paste (insert) before the scene node */
		} else {
			parent_node = (pfGroup *)scene_path->get(path_len-2);
			last_node_index = ((pfGroup *)parent_node)->searchChild(last_node);
			parent_node->insertChild(last_node_index, cut_node);
			last_node = modify_path(scene_path, 1, 1);

			display_debug2(pfPathString(scene_path, buffer, 32768));
			prev_last_node = NULL;	/* force display of the detail node info */
		}
	}

	/*** shift the current node to the left among its siblings ***/
	if (DEBOUNCE(key_shiftleft, prev_shiftleft)) {
		int	path_len;
		int	last_node_index;
		pfGroup	*parent_node;

		path_len = (int)scene_path->getNum();
		if (path_len < 2) {
			/* we can't shift the scene node */
		} else {
			parent_node = (pfGroup *)scene_path->get(path_len-2);
			last_node_index = ((pfGroup *)parent_node)->searchChild(last_node);

			if (last_node_index > 0) {
				parent_node->removeChild(last_node);
				parent_node->insertChild(last_node_index-1, last_node);

				display_debug2(pfPathString(scene_path, buffer, 32768));
				prev_last_node = NULL;	/* force display of the detail node info */
			}
		}
	}

	/*** shift the current node to the right among its siblings ***/
	if (DEBOUNCE(key_shiftright, prev_shiftright)) {
		int	path_len;
		int	last_node_index;
		pfGroup	*parent_node;

		path_len = (int)scene_path->getNum();
		if (path_len < 2) {
			/* we can't shift the scene node */
		} else {
			parent_node = (pfGroup *)scene_path->get(path_len-2);
			last_node_index = ((pfGroup *)parent_node)->searchChild(last_node);

			if (last_node_index < parent_node->getNumChildren()-1) {
				parent_node->removeChild(last_node);
#if 0 /* this isn't necessary if insertChild will allow 'insertion' after last child */
				if (last_node_index == parent_node->getNumChildren())
					parent_node->addChild(last_node);
				else	parent_node->insertChild(last_node_index+1, last_node);
#else
				parent_node->insertChild(last_node_index+1, last_node);
#endif

				display_debug2(pfPathString(scene_path, buffer, 32768));
				prev_last_node = NULL;	/* force display of the detail node info */
			}
		}
	}


	/******************************************************************************/
	/*** if the current node has changed, then update the detailed node display ***/
	if (prev_last_node != last_node || prev_skip_children != skip_children) {
		/*** we changed nodes ***/
		display_debug3(pfNodeString(last_node, NULL, 0, skip_children, buffer, 32768));
		display_debug3_top(pfNodeSiblingsString(last_node, 0, buffer, 2048));
	}
}

