/* ======================================================================
 *
 *  CCCCC    ++      ++   pfex0_purepf.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: December 13, 2000
 * CC   CC   ++      ++   Last Modified: December 13, 2000
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a Performer application program that */
/*   uses the only the SGI Performer library to handle the visual      */
/*   rendering.                                                        */
/*                                                                     */
/* The features of this example include:                               */
/*   - Initializing the Performer { system | library }                 */
/*   - Opening a Performer window                                      */
/*   - Creating a simple performer scenegraph                          */
/*   - The basic Performer Library functions:                          */
/*      * pfInit()                                                     */
/*      * pfSync()                                                     */
/*      * pfFrame()                                                    */
/*      * pfExit()                                                     */
/*                                                                     */
/* Considerations:                                                     */
/*   - extra effort in opening the GLX window                          */
/*   - ... getting inputs (eg. keyboard)                               */
/*                                                                     */
/* Functionality of this example:                                      */
/*   - Creates a single Scene Graph of a simple world, similar to the  */
/*      OpenGL example world.                                          */
/*                                                                     */
/* Copyright 2002, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */


/* This is very important -- these functions will be used as arguments. */
void	exit_application(int);

/* functions defined in this file */
void		update_world();
static pfScene	*createScene();


int main(int argc, char **argv)
{
	/**********************/
	/* setup the graphics */

	/* initialize Performer */
	pfInit();
	pfMultiprocess(PFMP_DEFAULT);

	/* configure Performer */
	pfConfig();


	/* create a channel for viewing the world */
	pfPipe *p = pfGetPipe(0);
	pfPipeWindow *pw = new pfPipeWindow(p);
	pw->setWinType(PFPWIN_TYPE_X);
	pw->setName("pfSimple");
	pw->setOriginSize(0, 224, 300, 300);
	pw->open();

	pfChannel *chan = new pfChannel(p);
	chan->setFOV(90.0f, 65.0);
	chan->setNearFar(1.0f, 8000.0f);


	/********************/
	/* create the world */

	/* Create a scene and assign it to the channels */
	chan->setScene(createScene());


	/***********************/
	/* do world simulation */
	/***********************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (1) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also using callback routines within    */
		/* the scene-graph to modify the world.                    */
		update_world();


		pfSync();
		/* latency critical operations go here */
		pfFrame();
	}

	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
	pfExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world()
{
	/* Nothing happens -- this is a static world. */

	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(float r, float g, float b);
pfNode *make_pyramid(float r, float g, float b);

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene()
{
	pfScene		*scene;
	pfGeoState	*gstate;
	pfSCS		*cs_node;
	pfMatrix	cs_matrix;

	scene = new pfScene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/**********************************************************/
	/***  Create an SCS (w/ matrix) for each static object  ***/

	/* a blue cube */
	cs_matrix.makeTrans(-2.0,  5.5, 0.5);
	cs_node = new pfSCS(cs_matrix);
	scene->addChild(cs_node);
	cs_node->addChild(make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));

	/* a red pyramid */
	cs_matrix.makeTrans( 2.0,  5.5, 0.0);
	cs_node = new pfSCS(cs_matrix);
	scene->addChild(cs_node);
	cs_node->addChild(make_pyramid(255.0/256.0, 100.0/256.0, 100.0/256.0));

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	/* NOTE, classes/structures that are specificed by the pr library     */
	/*   do not have "new" overloaded to use shared memory for allocating */
	/*   the memory, so we need to be specifically allocat them via the   */
	/*   CAVE's shared memory allocation API.                             */
	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}

/**************************/
/* draw_world(): does ... */
/**************************/
/*  NOTHING -- there is no application-specified world rendering in   */
/*    the Performer version of the CAVE library.  Performer does this */
/*    for us.                                                         */

