/* ======================================================================
 *
 *  CCCCC    ++      ++   pfex10_isect.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: March 7, 2000
 * CC   CC   ++      ++   Last Modified: October 30, 2003
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a Performer application program that */
/*   finds the nearest object pointed at by the wand and ...           */
/*                                                                     */
/* The new features of this example include:                           */
/*   Uses Performer's intersection testing to mark the pointed at object */
/*   - The basic Performer Library functions/features:                 */
/*	Intersection testing                                           */
/*	pfSeg/pfSegSet                                                 */
/*	pfHit                                                          */
/*      ...                                                            */
/* NOTE: objects can be selected, but they can't be grabbed.           */
/*                                                                     */
/* TODO: fix lighting problem when WIM is activated.(see pfex12b_geode)*/
/*                                                                     */
/* Copyright 2003, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

/* Performer library includes */
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfSwitch.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */

#include <Performer/pf/pfText.h>
#include <Performer/pr/pfString.h>
#include <Performer/pr/pfFont.h>

/* VR library includes */
#define PERFORMER
#include "freevr.h"
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */


/* Define a structure for information about each individual object. */
typedef struct {
		pfDCS		*dcs;		/* position of the object (controlled by a pfDCS) */
		pfVec4		*color;		/* vector that contains values used to color the object */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		object_count;
		int		selected_object;
		ObjectType	object[7];

		pfScene		*scene;		/* the entire graphical scene */
		pfGroup		*vw_node;	/*   just the virtual world   */
		pfGroup		*ui_node;	/*   just the user interface  */

		/* User interface control nodes */
		pfDCS		*wand_dcs;	/*   just the wand            */
		pfVec4		*wand_color;

		pfDCS		*wim_dcs;	/*   just the world-in-mini   */
		pfSwitch	*wim_switch;

		pfDCS		*marker_dcs;	/*   just the isect marker    */
		pfVec4		*marker_color;

		pfString	*info_text;
	} WorldDataType;

/* This is very important -- this function will be used as an argument. */
	void	exit_application(int);

/* functions defined in this file */
	void	update_world(WorldDataType *wd);
static	pfScene	*createScene(WorldDataType *wd);
static	int	travel_cb(pfTraverser *trav, void *);
static	int	object1_cb(pfTraverser *trav, void *);
static	int	object2_cb(pfTraverser *trav, void *);
static	int	wand_cb(pfTraverser *trav, void *);


int main(int argc, char **argv)
{
	WorldDataType	*wd;

	/* TODO: I do the signal setup a little different in Pf than OpenGL -- investigate */
	signal(SIGINT, SIG_IGN);

	/* initialize Performer */
	pfInit();

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, exit_application);
	vrConfigure(&argc, argv, NULL);

	/* configure Performer */
	pfConfig();


	/* create the channels for viewing the world */
	vrStart();

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   at the expense of some performance.                       */
  #if 0
	/* Turn on all possible statistics, except "fx pipe fill" */
	vrPfMasterChannel()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);
  #endif


	/********************/
	/* create the world */
	wd = (WorldDataType *)pfMalloc(sizeof(WorldDataType), pfGetSharedArena());

	/* Create a scene and assign it to the channels */
	vrPfMasterChannel()->setScene(createScene(wd));


	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (
		!vrGet2switchValue(0)
		) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also use callback routines within the  */
		/* scene-graph to modify the world.                        */
		update_world(wd);


		pfSync();
		/* latency critical operations go here */
		vrPfPreFrame();
		pfFrame();
		vrPfPostFrame();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
	vrExit();
	pfExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world(WorldDataType *wd)
{
static	vrMatrix	*mat_neg90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X, -90.0);	/* convert from Pf to OpenGL space */
static	vrMatrix	*mat_pos90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X,  90.0);	/* convert from OpenGL to Pf space  (ogl2pf) */
	vrVector	wand_pointvec;
	vrPoint		wand_point;
#define wand_location wand_point.v
	pfVec3		start_point;			/* start point of intersection segment */
	pfVec3		end_point;			/* end point of intersection segment */
	pfMatrix	obj_mat;
	int		wim_button;
	pfMatrix	tmpmat1, tmpmat2;		/* These are used by direct method of wand avatar positioning. */
	/* These varialbes are used for the doing intersection testing */
static	pfSegSet	*segset = NULL;			/* list of line segments to test against (man pfSeg) */
	pfHit		**hits[32];			/* list of intersections detected */
	int		isect;				/* boolean indicating intersecting found */

	/********************************************/
	/* initialize the intersection testing info */
	if (segset == NULL) {		/* only need to initialize once */
		segset = new pfSegSet;
		segset->activeMask = 1;
		segset->isectMask = 0xFFFF;
		segset->discFunc = NULL;
		segset->bound = NULL;
		segset->mode = PFTRAV_IS_PRIM	/* check isects with quads or trangles */
			| PFTRAV_IS_CULL_BACK;	/* ignore back-facing polygons*/

		/* this sets an initial segment, but will be overwritten */
		/*   before we do the first intersection test.           */
		segset->segs[0].dir.set(0.0, 0.0, -1.0);
		segset->segs[0].length = 50000.0;
	}

	/*************************************************************/
	/* do intersection testing using current wand position in VW */
	vrPointGetVWFromUser6sensor(&wand_point, 0, WAND_SENSOR);
	vrPointTransformByMatrix(&wand_point, &wand_point, mat_pos90x);		/* adjust to put into Performer space */

	vrVectorGetVWFromUser6sensorDir(&wand_pointvec, 0, WAND_SENSOR, VRDIR_FORE);
	vrVectorTransformByMatrix(&wand_pointvec, &wand_pointvec, mat_pos90x);		/* adjust to put into Performer space */
	start_point = pfVec3(wand_location[0], wand_location[1], wand_location[2]);
	end_point = start_point + (pfVec3(wand_pointvec.v[0], wand_pointvec.v[1], wand_pointvec.v[2]) * 50000.0);
	segset->segs[0].makePts(start_point, end_point);
	isect = wd->vw_node->isect(segset, hits);

	if (isect) {
		pfVec3	 isect_obj_location;	/* where intersected on the object */
		pfVec3	 isect_world_location;	/* where in the world intersected */
		pfMatrix isect_world_xform;	/* xform matrix at intersection */

		wd->info_text->setString("hit something");

		/* get where intersected with object (in it's coord-sys) */
		(*hits[0])->query(PFQHIT_POINT, &isect_obj_location);

		/* convert to world coord-sys */
		(*hits[0])->query(PFQHIT_XFORM, &isect_world_xform);
		isect_world_location.xformPt(isect_obj_location, isect_world_xform);

		/* place a marker (sphere) on the object */
		wd->marker_dcs->setTrans(isect_world_location[PF_X], isect_world_location[PF_Y], isect_world_location[PF_Z]);
	} else {
		wd->info_text->setString("no hit");
	}


	/*************************************************/
	/* turn on/off w-i-m based on middle wand button */
	wim_button = vrGet2switchValue(2);

	wd->wim_switch->setVal((wim_button ? PFSWITCH_ON : PFSWITCH_OFF));


	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}


/**********************************************************************/
static int travel_cb(pfTraverser *trav, void *)
{
#define SPEED 0.02f
static	double		last_time = -1.0;
	double		this_time = pfGetTime();
	double		delta_time;
	float		delta_move;
	pfDCS   	*travel_dcs = (pfDCS *)trav->getNode();  /* Get a pointer to the DCS node itself */
	float		joy_x, joy_y;
	vrVector	wand_pointvec;

	/***************************************************/
	/* Determine delta time from last simulation frame */
	if (last_time == -1.0)
		last_time = this_time;

	delta_time = this_time - last_time;

	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	/* notice rotation for Performer is about the z-axis instead of y */
	if (fabs(joy_x) > 0.125)
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (joy_x * delta_time * SPEED * -7.5));

	if (fabs(joy_y) > 0.125) {
		delta_move = joy_y * delta_time * SPEED;
		vrVectorGetRWFrom6sensorDir(&wand_pointvec, WAND_SENSOR, VRDIR_FORE);
		vrUserTravelTranslate3d(VR_ALLUSERS,
			wand_pointvec.v[VR_X] * delta_move,
			wand_pointvec.v[VR_Y] * delta_move,
			wand_pointvec.v[VR_Z] * delta_move);
	}

	/* Reset travel on button 3 */
	if (vrGet2switchValue(3))
		vrUserTravelReset(VR_ALLUSERS);

	/* Load the DCS with the latest navigation matrix */
	/* NOTE: In FreeVR, because each user can have their own travel CS,   */
	/*   the pfDCS cannot be changed in the APP callback.  So the AP must */
	/*   assign the special callback function to the travel node as a     */
	/*   CULL callback.  This function is vrPfDCSTransformUserTravelCB(). */

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}


/**********************************************************************/
static int object1_cb(pfTraverser *trav, void *)
{
	float		obj_x, obj_y, obj_z;
	float		sim_time = pfGetFrameTimeStamp();
	pfDCS		*objectDcs = (pfDCS *) trav->getNode();

	obj_z =  5.5;
	obj_x =  3.0*sin(sim_time * 2.0);
	obj_y = -3.0*cos(sim_time * 2.0) + 4.0;
	objectDcs->setTrans(obj_x, obj_y, obj_z);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}


/**********************************************************************/
static int object2_cb(pfTraverser *trav, void *)
{
	float		obj_x, obj_y, obj_z;
	float		sim_time = pfGetFrameTimeStamp();
	pfDCS		*objectDcs = (pfDCS *) trav->getNode();

	obj_y =  4.0;
	obj_x =  3.0*cos(sim_time * 2.0);
	obj_z =  3.0*sin(sim_time * 2.0) + 5.5;
	objectDcs->setTrans(obj_x, obj_y, obj_z);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}


/**********************************************************************/
static int wand_cb(pfTraverser *trav, void *data)
{
	pfMatrix	tmpmat1, tmpmat2;
	pfDCS		*sensor_dcs = (pfDCS *)trav->getNode();
	int		sensor_num = *((int *)data);		/* number of the 6-dof sensor tied to this node */

	vrPfDCSTransform6sensor(sensor_dcs, sensor_num);

	/* TODO: in FreeVR, the scale is currently not maintained, so need to restore */
	if (sensor_num == 1) {
		/* rotate/scale the DCS so the (small) wand points forward */
		sensor_dcs->setScale(  0.1, 0.5, 0.1);
		tmpmat1.makeRot(-90.0, 1.0, 0.0, 0.0);
		sensor_dcs->getMat(tmpmat2);	/* WARNING: unobvious call-by-ref */
		tmpmat2.preMult(tmpmat1);
		sensor_dcs->setMat(tmpmat2);	/* WARNING: unobvious call-by-ref */
		sensor_dcs->setScale(0.25, 0.25, 0.25);	/* pfBUG: for some reason, the middle value has no effect, but the value from above does. */
	}
	if (sensor_num == 2)
		sensor_dcs->setScale( 0.20, 0.20, 0.20);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(pfVec4 *color_vec);
pfNode *make_pyramid(pfVec4 *color_vec);
pfNode *make_marker(pfVec4 *color_vec);

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene(WorldDataType *wd)
{
	int		*int_data;	/* used for setting callback data */
	pfScene		*scene;
	pfGeoState	*gstate;
	pfVec4		*color;
	pfDCS		*cs_node;
	pfDCS		*travel_cs;

	wd->object_count = 0;
	wd->selected_object = -1;

	/* find Performer data objects in a common directory */
	pfFilePath(".:/afs/ncsa/projects/CS397WRS/PFdata");

	scene = new pfScene;
	wd->scene = scene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/* create generic nodes for the virtual world and UI objects */
	wd->vw_node = new pfGroup;
	wd->ui_node = new pfGroup;

	/* the ui_node goes directly under the scene, the vw_node comes next */
	scene->addChild(wd->ui_node);


	/***************************************/
	/*** Create a DCS for travel control ***/
	travel_cs = new pfDCS();
	scene->addChild(travel_cs);

	/* use a callback to adjust the navigation */
	travel_cs->setTravFuncs(PFTRAV_APP, travel_cb, NULL);
	travel_cs->setTravFuncs(PFTRAV_CULL, vrPfDCSTransformUserTravelCB, NULL);
	/* put the generic VW node under the travel DCS */
	travel_cs->addChild(wd->vw_node);


	/**********************************************************/
	/*** Create a DCS (w/ matrix) for each *dynamic* object ***/

	/* a blue floor (ie. large flat cube) */
	cs_node = new pfDCS();
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans(0.0, 0.0, 0.0);
	cs_node->setScale(10.0, 10.0, 0.1);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(25.0/256.0, 25.0/256.0, 200.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));

	/* a blue cube */
	cs_node = new pfDCS();
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans(-2.0, 4.0, 3.5);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));
	cs_node->setTravFuncs(PFTRAV_APP, object1_cb, NULL);	/* use a callback to relocate this object */
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a red pyramid */
	cs_node = new pfDCS();
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans( 2.0, 4.0, 3.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(255.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));
	cs_node->setTravFuncs(PFTRAV_APP, object2_cb, NULL);	/* use a callback to relocate this object */
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a green upside down pyramid */
	cs_node = new pfDCS();
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans( 0.0, 4.0, 4.0);
	cs_node->setScale(1.0, 1.0, -1.0);	/* NOTE: flip in Z, not Y */
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 255.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a car model, loaded from a file */
	cs_node = new pfDCS();
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans( 0.0, 4.0, 0.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(pfdLoadFile("esprit.pfb"));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;


	/* set the entire world for isect testing (exceptions to come later) */
	scene->setTravMask(PFTRAV_ISECT, 0x1, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	/***************************************************/
	/*** Create nodes for each user interface object ***/

	/* a purple cube -- in the RWcs not the VWcs */
	cs_node = new pfDCS();
	wd->ui_node->addChild(cs_node);	/* NOTE: attatched directly to the UI node */
	cs_node->setTrans(-4.0, -1.0, 0.5);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(200.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));
	cs_node->setTravMask(PFTRAV_ISECT, 0x0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	/* some white text -- rendered in the RWcs */
	cs_node = new pfDCS();
	wd->ui_node->addChild(cs_node);	/* NOTE: attatched directly to the UI node */
	cs_node->setTrans(-3.0, 5.0, 2.0);
	pfText *text_node = new pfText;
	cs_node->addChild(text_node);
	cs_node->setTravMask(PFTRAV_ISECT, 0x0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	wd->info_text = new pfString;
	text_node->addString(wd->info_text);
	wd->info_text->setFont(pfdLoadFont_type1("Times-Elfin", PFDFONT_FILLED));
	wd->info_text->setMode(PFSTR_JUSTIFY,PFSTR_LEFT);
	wd->info_text->setColor(1.0, 1.0, 1.0, 1.0);
	wd->info_text->setString("-- info text --");


	/* the Wand -- a long cyan pyramid */
	wd->wand_dcs = new pfDCS();
	scene->addChild(wd->wand_dcs);	/* NOTE: the wand is still attached to the scene */
	wd->wand_dcs->setRot(0.0, -90.0, 0.0);
	wd->wand_dcs->setTrans( 0.0, 0.0, 0.0);
	wd->wand_dcs->setScale( 0.5, 1.0, 0.5);
	wd->wand_color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	wd->wand_color->set(100.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	wd->wand_dcs->addChild(make_pyramid(wd->wand_color));

	/* use a callback to relocate the wand */
	int_data = (int *)pfMalloc(sizeof(int), pfGetSharedArena());
	*int_data = 1;
	wd->wand_dcs->setTravFuncs(PFTRAV_APP, wand_cb, NULL);
	wd->wand_dcs->setTravData(PFTRAV_APP, (void *)int_data);

	/* the World-in-Miniature node */
	wd->wim_switch = new pfSwitch;
	wd->ui_node->addChild(wd->wim_switch);
	wd->wim_switch->setVal(PFSWITCH_ON);

	wd->wim_dcs = new pfDCS;
	wd->wim_switch->addChild(wd->wim_dcs);
	wd->wim_dcs->setRot(90.0, 0.0, 0.0);
	wd->wim_dcs->setTrans( 0.0, 0.0, 0.0);
	wd->wim_dcs->setScale( 0.20, 0.20, 0.20);
	wd->wim_dcs->addChild(wd->vw_node);	/* the VW is the WIM */

	/* use same wand callback to relocate the wim-wand */
	int_data = (int *)pfMalloc(sizeof(int), pfGetSharedArena());
	*int_data = 2;
	wd->wim_dcs->setTravFuncs(PFTRAV_APP, wand_cb, NULL);
	wd->wim_dcs->setTravData(PFTRAV_APP, (void *)int_data);

	/* TODO: need to find out if it's possible to make all LODs be low-res*/

	/* the intersection marker node */
		/* needs to be in the virtual world? why??     */
		/* I need to allow isect w/ the purple box and */
		/* see how that works.                         */
	wd->marker_dcs = new pfDCS;
	wd->vw_node->addChild(wd->marker_dcs);
	wd->marker_dcs->setRot(0.0, 0.0, 0.0);
	wd->marker_dcs->setTrans( 0.0, 0.0, 0.0);
	wd->marker_dcs->setScale( 0.20, 0.20, 0.20);
	wd->marker_color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	wd->marker_color->set(100.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	wd->marker_dcs->addChild(make_marker(wd->marker_color));

	/* don't intersection test with the marker itself */
	wd->marker_dcs->setTravMask(PFTRAV_ISECT, 0x0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/************************************************************/
/* make a marker to indicate where an intersection occures. */
pfNode *make_marker(pfVec4 *color_vec)
{
#define SPHERE_POLYS	25
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewSphere(SPHERE_POLYS, pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}

