/* ======================================================================
 *
 *  CCCCC    ++      ++   pfex12_geode.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: December 6, 2000
 * CC   CC   ++      ++   Last Modified: October 30, 2003
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a Performer application program that */
/*   adds a texture map to the hand-made pfGeode of example 12.        */
/*                                                                     */
/* The new features of this example include:                           */
/*   - The basic Performer Library functions/features:                 */
/*      ...                                                            */
/*   - The basic vrPF Library functions:                               */
/*      [none introduced in this example]                              */
/*                                                                     */
/* NOTE: objects can be selected, but they can't be grabbed.           */
/*                                                                     */
/* Copyright 2003, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

/* Performer library includes */
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfSwitch.h>
#include <Performer/pf/pfTraverser.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */

#include <Performer/pf/pfText.h>
#include <Performer/pr/pfString.h>
#include <Performer/pr/pfFont.h>

#include <Performer/pr/pfTexture.h>

/* VR library includes */
#define PERFORMER
#include "freevr.h"
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */

/* pfdebug.c++ functions */
int	pfdebug_initialize_display(char *tty);
void	pfdebug_terminate_display();
void	pfdebug_display_sg(pfScene *scene);


/* Define a structure for information about each individual object. */
typedef struct {
		pfDCS		*dcs;		/* position of the object (controlled by a pfDCS) */
		pfVec4		*color;		/* vector that contains values used to color the object */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		object_count;
		int		selected_object;
		ObjectType	object[32];
		int		sg_view;

		pfScene		*scene;		/* the entire graphical scene */
		pfGroup		*vw_node;	/*   just the virtual world   */
		pfGroup		*ui_node;	/*   just the user interface  */

		/* User interface control nodes */
		pfDCS		*wand_dcs;	/*   just the wand            */
		pfVec4		*wand_color;

		pfDCS		*wim_dcs;	/*   just the world-in-mini   */
		pfSwitch	*wim_switch;

		pfDCS		*marker_dcs;	/*   just the isect marker    */
		pfSwitch	*marker_switch;	/*   switch node parent of the marker */
		pfVec4		*marker_color;

		pfString	*info_text;
	} WorldDataType;

/* This is very important -- this function will be used as an argument. */
	void	exit_application(int);

/* functions defined in this file */
	void	update_world(WorldDataType *wd);
static	pfScene	*createScene(WorldDataType *wd);
static	int	travel_cb(pfTraverser *trav, void *);
static	int	object1_cb(pfTraverser *trav, void *);
static	int	object2_cb(pfTraverser *trav, void *);
static	int	wand_cb(pfTraverser *trav, void *);


int main(int argc, char **argv)
{
	WorldDataType	*wd;

	/* TODO: I do the signal setup a little different in Pf than OpenGL -- investigate */
	signal(SIGINT, SIG_IGN);

	/* initialize Performer */
	pfInit();

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, exit_application);
	vrConfigure(&argc, argv, NULL);

	/* configure Performer */
	pfConfig();


	/* create the channels for viewing the world */
	vrStart();

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   at the expense of some performance.                       */
  #if 0
	/* Turn on all possible statistics, except "fx pipe fill" */
	vrPfMasterChannel()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);
  #endif


	/********************/
	/* create the world */
	wd = (WorldDataType *)pfMalloc(sizeof(WorldDataType), pfGetSharedArena());

	/* Create a scene and assign it to the channels */
	vrPfMasterChannel()->setScene(createScene(wd));

	/* initialize the scenegraph text display */
	wd->sg_view = pfdebug_initialize_display(getenv("DBGTTY"));


	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (
		!vrGet2switchValue(0)
		) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also use callback routines within the  */
		/* scene-graph to modify the world.                        */
		update_world(wd);

		/*********************************************/
		/* display text view of performer scenegraph */
		if (wd->sg_view)
			pfdebug_display_sg(wd->scene);

		pfSync();
		/* latency critical operations go here */
		vrPfPreFrame();
		pfFrame();
		vrPfPostFrame();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int signal)
{
	pfdebug_terminate_display();

	vrExit();
	pfExit();

	printf("We never get this far.\n");
	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world(WorldDataType *wd)
{
static	vrMatrix	*mat_neg90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X, -90.0);	/* convert from Pf to OpenGL space */
static	vrMatrix	*mat_pos90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X,  90.0);	/* convert from OpenGL to Pf space  (ogl2pf) */
	vrVector	wand_pointvec;
	vrPoint		wand_point;
#define wand_location wand_point.v
	pfVec3		start_point;			/* start point of intersection segment */
	pfVec3		end_point;			/* end point of intersection segment */
	pfMatrix	obj_mat;
	int		prev_selected_object;
	int		wim_button;
	pfMatrix	tmpmat1, tmpmat2;		/* These are used by direct method of wand avatar positioning. */
	/* These varialbes are used for the doing intersection testing */
static	pfSegSet	*segset = NULL;			/* list of line segments to test against (man pfSeg) */
	pfHit		**hits[32];			/* list of intersections detected */
	int		isect;				/* boolean indicating intersecting found */

	/********************************************/
	/* initialize the intersection testing info */
	if (segset == NULL) {		/* only need to initialize once */
		segset = new pfSegSet;
		segset->activeMask = 1;
		segset->isectMask = 0xFFFF;
		segset->discFunc = NULL;
		segset->bound = NULL;
		segset->mode = PFTRAV_IS_PRIM	/* check isects with quads or trangles */
			| PFTRAV_IS_NORM	/* return normals */
			| PFTRAV_IS_PATH	/* store traversal path info */
			| PFTRAV_IS_CULL_BACK;	/* ignore back-facing polygons*/

		/* this sets an initial segment, but will be overwritten */
		/*   before we do the first intersection test.           */
		segset->segs[0].dir.set(0.0, 0.0, -1.0);
		segset->segs[0].length = 50000.0;
	}

	/*************************************************************/
	/* do intersection testing using current wand position in VW */
	vrPointGetVWFromUser6sensor(&wand_point, 0, WAND_SENSOR);
	vrPointTransformByMatrix(&wand_point, &wand_point, mat_pos90x);		/* adjust to put into Performer space */

	vrVectorGetVWFromUser6sensorDir(&wand_pointvec, 0, WAND_SENSOR, VRDIR_FORE);
	vrVectorTransformByMatrix(&wand_pointvec, &wand_pointvec, mat_pos90x);		/* adjust to put into Performer space */
	start_point = pfVec3(wand_location[0], wand_location[1], wand_location[2]);
	end_point = start_point + (pfVec3(wand_pointvec.v[0], wand_pointvec.v[1], wand_pointvec.v[2]) * 50000.0);
	segset->segs[0].makePts(start_point, end_point);
	isect = wd->vw_node->isect(segset, hits);

	prev_selected_object = wd->selected_object;

	if (isect) {
		pfVec3	 isect_obj_location;	/* where intersected on the object */
		pfVec3	 isect_world_location;	/* where in the world intersected */
		pfMatrix isect_world_xform;	/* xform matrix at intersection */
		pfNode	 *isect_node;
		pfNode	 *isect_parent;
		int	 count;

		wd->info_text->setString("hit something");

		/* get where intersected with object (in it's coord-sys) */
		(*hits[0])->query(PFQHIT_POINT, &isect_obj_location);

		/* convert to world coord-sys */
		(*hits[0])->query(PFQHIT_XFORM, &isect_world_xform);
		isect_world_location.xformPt(isect_obj_location, isect_world_xform);

		/* place a marker (sphere) on the object */
		wd->marker_dcs->setTrans(isect_world_location[PF_X], isect_world_location[PF_Y], isect_world_location[PF_Z]);

		wd->marker_switch->setVal(PFSWITCH_ON);

		/* find out with what object we are intersecting */
		(*hits[0])->query(PFQHIT_NODE, &isect_node);
#undef DEBUG_ISECT
#ifdef DEBUG_ISECT
		printf("intersecting with node %p, child of %p\n", isect_node, isect_node->getParent(0));
#endif

		/* find what object we are intersecting with by comparing w/ parent */
		isect_parent = isect_node->getParent(0);
		wd->selected_object = -1;
		for (count = 0; count < wd->object_count; count++) {
			if (isect_parent == wd->object[count].dcs)
				wd->selected_object = count;
		}

	} else {
		wd->info_text->setString("no hit");
		wd->marker_switch->setVal(PFSWITCH_OFF);
		wd->selected_object = -1;
	}

	/* adjust object colors if selected object just changed */
	if (prev_selected_object != wd->selected_object) {
		if (prev_selected_object != -1) {
			/* restore the previosly selected object */
			*(wd->object[prev_selected_object].color) *= 2.0;
		}
		if (wd->selected_object != -1) {
			/* enhance the currently selected object */
			*(wd->object[wd->selected_object].color) *= 0.5;
		}
	}


	/*************************************************/
	/* turn on/off w-i-m based on middle wand button */
	wim_button = vrGet2switchValue(2);

	wd->wim_switch->setVal((wim_button ? PFSWITCH_ON : PFSWITCH_OFF));


	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}


/**********************************************************************/
static int travel_cb(pfTraverser *trav, void *)
{
#define SPEED 0.005f
static	double		last_time = -1.0;
	double		this_time = pfGetTime();
	double		delta_time;
	float		delta_move;
	pfDCS		*travel_dcs = (pfDCS *)trav->getNode();  /* Get a pointer to the DCS node itself */
	float		joy_x, joy_y;
	vrVector	wand_pointvec;

	/***************************************************/
	/* Determine delta time from last simulation frame */
	if (last_time == -1.0)
		last_time = this_time;

	delta_time = this_time - last_time;

	joy_x = vrGetValuatorValue(0);
	joy_y = vrGetValuatorValue(1);

	/* notice rotation for Performer is about the z-axis instead of y */
	if (fabs(joy_x) > 0.125)
		vrUserTravelRotateId(VR_ALLUSERS, VR_Y, (joy_x * delta_time * SPEED * -7.5));

	if (fabs(joy_y) > 0.125) {
		delta_move = joy_y * delta_time * SPEED;
		vrVectorGetRWFrom6sensorDir(&wand_pointvec, WAND_SENSOR, VRDIR_FORE);
		vrUserTravelTranslate3d(VR_ALLUSERS,
			wand_pointvec.v[VR_X] * delta_move,
			wand_pointvec.v[VR_Y] * delta_move,
			wand_pointvec.v[VR_Z] * delta_move);
	}

	/* Reset travel on button 3 */
	if (vrGet2switchValue(3))
		vrUserTravelReset(VR_ALLUSERS);

	/* Load the DCS with the latest navigation matrix */
	/* NOTE: In FreeVR, because each user can have their own travel CS,   */
	/*   the pfDCS cannot be changed in the APP callback.  So the AP must */
	/*   assign the special callback function to the travel node as a     */
	/*   CULL callback.  This function is vrPfDCSTransformUserTravelCB(). */

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}


/**********************************************************************/
static int object1_cb(pfTraverser *trav, void *)
{
	float		obj_x, obj_y, obj_z;
	float		sim_time = pfGetFrameTimeStamp();
	pfDCS		*objectDcs = (pfDCS *) trav->getNode();

	obj_z =  5.5;
	obj_x =  3.0*sin(sim_time * 2.0);
	obj_y = -3.0*cos(sim_time * 2.0) + 4.0;
	objectDcs->setTrans(obj_x, obj_y, obj_z);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}


/**********************************************************************/
static int object2_cb(pfTraverser *trav, void *)
{
	float		obj_x, obj_y, obj_z;
	float		sim_time = pfGetFrameTimeStamp();
	pfDCS		*objectDcs = (pfDCS *) trav->getNode();

	obj_y =  4.0;
	obj_x =  3.0*cos(sim_time * 2.0);
	obj_z =  3.0*sin(sim_time * 2.0) + 5.5;
	objectDcs->setTrans(obj_x, obj_y, obj_z);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}


/**********************************************************************/
static int wand_cb(pfTraverser *trav, void *data)
{
	pfMatrix	tmpmat1, tmpmat2;
	pfDCS		*sensor_dcs = (pfDCS *)trav->getNode();
	int		sensor_num = *((int *)data);		/* number of the 6-dof sensor tied to this node */

	vrPfDCSTransform6sensor(sensor_dcs, sensor_num);

	/* TODO: in FreeVR, the scale is currently not maintained, so need to restore */
	if (sensor_num == 1) {
		/* rotate/scale the DCS so the (small) wand points forward */
		sensor_dcs->setScale(  0.1, 0.5, 0.1);
		tmpmat1.makeRot(-90.0, 1.0, 0.0, 0.0);
		sensor_dcs->getMat(tmpmat2);	/* WARNING: unobvious call-by-ref */
		tmpmat2.preMult(tmpmat1);
		sensor_dcs->setMat(tmpmat2);	/* WARNING: unobvious call-by-ref */
		sensor_dcs->setScale(0.25, 0.25, 0.25);	/* pfBUG: for some reason, the middle value has no effect, but the value from above does. */
	}
	if (sensor_num == 2)
		sensor_dcs->setScale( 0.20, 0.20, 0.20);

	/* Tell Performer to continue traversing */
	return PFTRAV_CONT;
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(char *name, pfVec4 *color_vec);
pfNode *make_cube_byhand(char *name, pfVec4 *color_vec);
pfNode *make_pyramid(char *name, pfVec4 *color_vec);
pfNode *make_marker(char *name, pfVec4 *color_vec);

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene(WorldDataType *wd)
{
	int		*int_data;	/* used for setting callback data */
	pfScene		*scene;
	pfGeoState	*gstate;
	pfVec4		*color;
	pfDCS		*cs_node;
	pfDCS		*travel_cs;

	wd->object_count = 0;
	wd->selected_object = -1;

	/* find Performer data objects in a common directory */
	pfFilePath(".:/afs/ncsa/projects/CS397WRS/PFdata");

	scene = new pfScene;
	scene->setName("Top of the world");
	wd->scene = scene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/* create generic nodes for the virtual world and UI objects */
	wd->vw_node = new pfGroup;
	wd->vw_node->setName("virtual world group");
	wd->ui_node = new pfGroup;
	wd->ui_node->setName("user interface group");

	/* the ui_node goes directly under the scene, the vw_node comes next */
	scene->addChild(wd->ui_node);


	/***************************************/
	/*** Create a DCS for travel control ***/
	travel_cs = new pfDCS();
	travel_cs->setName("travel DCS");
	scene->addChild(travel_cs);

	/* use a callback to adjust the navigation */
	travel_cs->setTravFuncs(PFTRAV_APP, travel_cb, NULL);
	travel_cs->setTravFuncs(PFTRAV_CULL, vrPfDCSTransformUserTravelCB, NULL);
	/* put the generic VW node under the travel DCS */
	travel_cs->addChild(wd->vw_node);


	/**********************************************************/
	/*** Create a DCS (w/ matrix) for each *dynamic* object ***/

	/* a blue floor (ie. large flat cube) */
	cs_node = new pfDCS();
	cs_node->setName("Floor DCS");
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans(0.0, 0.0, 0.0);
	cs_node->setScale(10.0, 10.0, 0.1);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(25.0/256.0, 25.0/256.0, 200.0/256.0, 1.0);
	cs_node->addChild(make_cube("Floor", color));

	/* a blue cube */
	cs_node = new pfDCS();
	cs_node->setName("Blue Cube DCS");
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans(-2.0, 4.0, 3.5);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube_byhand("Blue Cube", color));
	cs_node->setTravFuncs(PFTRAV_APP, object1_cb, NULL);	/* use a callback to relocate this object */
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a red pyramid */
	cs_node = new pfDCS();
	cs_node->setName("Red Pyramid DCS");
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans( 2.0, 4.0, 3.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(255.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid("Red Pyramid", color));
	cs_node->setTravFuncs(PFTRAV_APP, object2_cb, NULL);	/* use a callback to relocate this object */
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a green upside down pyramid */
	cs_node = new pfDCS();
	cs_node->setName("Green Pyramid DCS");
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans( 0.0, 4.0, 4.0);
	cs_node->setScale(1.0, 1.0, -1.0);	/* NOTE: flip in Z, not Y */
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 255.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid("Green Pyramid", color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

#if 0 /* disable the car for this example */
	/* a car model, loaded from a file */
	cs_node = new pfDCS();
	cs_node->setName("Car Model DCS");
	wd->vw_node->addChild(cs_node);
	cs_node->setTrans( 0.0, 4.0, 0.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(pfdLoadFile("esprit.pfb"));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;
#endif


	/* set the entire world for isect testing (exceptions to come later) */
	scene->setTravMask(PFTRAV_ISECT, 0x1, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	/***************************************************/
	/*** Create nodes for each user interface object ***/

	/* a purple cube -- in the RWcs not the VWcs */
	cs_node = new pfDCS();
	cs_node->setName("Purple Cube DCS");
	wd->ui_node->addChild(cs_node);	/* NOTE: attatched directly to the UI node */
	cs_node->setTrans(-4.0, -1.0, 0.5);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(200.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube("Purple Cube", color));
	cs_node->setTravMask(PFTRAV_ISECT, 0x0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	/* some white text -- rendered in the RWcs */
	cs_node = new pfDCS();
	cs_node->setName("Text DCS");
	wd->ui_node->addChild(cs_node);	/* NOTE: attatched directly to the UI node */
	cs_node->setTrans(-3.0, 5.0, 2.0);
	pfText *text_node = new pfText;
	text_node->setName("Text Message");
	cs_node->addChild(text_node);
	cs_node->setTravMask(PFTRAV_ISECT, 0x0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	wd->info_text = new pfString;
	text_node->addString(wd->info_text);
	wd->info_text->setFont(pfdLoadFont_type1("Times-Elfin", PFDFONT_FILLED));
	wd->info_text->setMode(PFSTR_JUSTIFY,PFSTR_LEFT);
	wd->info_text->setColor(1.0, 1.0, 1.0, 1.0);
	wd->info_text->setString("-- info text --");


	/* the Wand -- a long cyan pyramid */
	wd->wand_dcs = new pfDCS();
	wd->wand_dcs->setName("Wand DCS");
	scene->addChild(wd->wand_dcs);	/* NOTE: the wand is still attached to the scene */
	wd->wand_dcs->setRot(0.0, -90.0, 0.0);
	wd->wand_dcs->setTrans( 0.0, 0.0, 0.0);
	wd->wand_dcs->setScale( 0.1, 0.1, 1.5);
	wd->wand_color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	wd->wand_color->set(100.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	wd->wand_dcs->addChild(make_pyramid("Wand", wd->wand_color));

	/* use a callback to relocate the wand */
	int_data = (int *)pfMalloc(sizeof(int), pfGetSharedArena());
	*int_data = 1;
	wd->wand_dcs->setTravFuncs(PFTRAV_APP, wand_cb, NULL);
	wd->wand_dcs->setTravData(PFTRAV_APP, (void *)int_data);

	/* the World-in-Miniature node */
	wd->wim_switch = new pfSwitch;
	wd->wim_switch->setName("WIM Switch");
	wd->ui_node->addChild(wd->wim_switch);
	wd->wim_switch->setVal(PFSWITCH_ON);

	wd->wim_dcs = new pfDCS;
	wd->wim_dcs->setName("WIM DCS");
	wd->wim_switch->addChild(wd->wim_dcs);
	wd->wim_dcs->setRot(90.0, 0.0, 0.0);
	wd->wim_dcs->setTrans( 0.0, 0.0, 0.0);
	wd->wim_dcs->setScale( 0.20, 0.20, 0.20);
	wd->wim_dcs->addChild(wd->vw_node);	/* the VW is the WIM */

	/* use same wand callback to relocate the wim-wand */
	int_data = (int *)pfMalloc(sizeof(int), pfGetSharedArena());
	*int_data = 2;
	wd->wim_dcs->setTravFuncs(PFTRAV_APP, wand_cb, NULL);
	wd->wim_dcs->setTravData(PFTRAV_APP, (void *)int_data);

	/* TODO: need to find out if it's possible to make all LODs be low-res*/

	/* the intersection marker node */
		/* needs to be in the virtual world? why??     */
		/* I need to allow isect w/ the purple box and */
		/* see how that works.                         */
	wd->marker_switch = new pfSwitch;
	wd->marker_switch->setName("Marker Switch");
	wd->marker_switch->setVal(PFSWITCH_ON);
	wd->vw_node->addChild(wd->marker_switch);

	wd->marker_dcs = new pfDCS;
	wd->marker_dcs->setName("Marker DCS");
	wd->marker_switch->addChild(wd->marker_dcs);
	wd->marker_dcs->setRot(0.0, 0.0, 0.0);
	wd->marker_dcs->setTrans( 0.0, 0.0, 0.0);
	wd->marker_dcs->setScale( 0.20, 0.20, 0.20);
	wd->marker_color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	wd->marker_color->set(100.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	wd->marker_dcs->addChild(make_marker("I-sect Marker", wd->marker_color));

	/* don't intersection test with the marker itself */
	wd->marker_dcs->setTravMask(PFTRAV_ISECT, 0x0, PFTRAV_SELF | PFTRAV_DESCEND, PF_SET);

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(char *name, pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, NULL);
	new_geode = new pfGeode;
	new_geode->setName(name);
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/****************************************/
/* make a unit cube of the given color without the aid of the pfd library. */
/*   Note that this cube is the same as the shapes.c cube. */
pfNode *make_cube_byhand(char *name, pfVec4 *color_vec)
{
static	float cube_vertices[8][3] = {		/* specifies the (x,y,z)   */
		{ -1.0,  1.0,  1.0 },	/* 0 */	/* location of each vertex */
		{ -1.0, -1.0,  1.0 },	/* 1 */
		{  1.0, -1.0,  1.0 },	/* 2 */
		{  1.0,  1.0,  1.0 },	/* 3 */
		{  1.0,  1.0, -1.0 },	/* 4 */
		{  1.0, -1.0, -1.0 },	/* 5 */
		{ -1.0, -1.0, -1.0 },	/* 6 */
		{ -1.0,  1.0, -1.0 },	/* 7 */
	};

static	unsigned short cube_polygons[6][4] = {	/* specifices the vertices */
		{ 0, 1, 2, 3 },	/* front  */	/* of each polygon         */
		{ 3, 2, 5, 4 },	/* right  */
		{ 5, 6, 7, 4 },	/* back   */
		{ 7, 6, 1, 0 },	/* left   */
		{ 6, 5, 2, 1 },	/* bottom */
		{ 0, 3, 4, 7 },	/* top    */
	};

static	float cube_normals[6][3] = {		/* specifies the normal */
		{  0.0,  0.0,  1.0 }, /* front  ** of each polygon      */
		{  1.0,  0.0,  0.0 }, /* right  */
		{  0.0,  0.0, -1.0 }, /* back   */
		{ -1.0,  0.0,  0.0 }, /* left   */
		{  0.0, -1.0,  0.0 }, /* bottom */
		{  0.0,  1.0,  0.0 }, /* top    */
	};

static	int	poly_lengths[8] = { 4, 4, 4, 4, 4, 4, 4, 4 };
static	int	sequential_index[6] = { 0, 1, 2, 3, 4, 5 };

	pfVec3		*verts;
	pfVec3		*norms;
	int		*lengths;
	unsigned short	*index;
	unsigned short	*sindex;
	pfGeoSet	*geoset;
	pfGeoState	*geostate;
	pfGeode		*new_geode;

	/* make a generic geostate */
	geostate = new pfGeoState;
	geostate->makeBasic();
	geostate->setMode(PFSTATE_ENLIGHTING, PF_ON);
#if 0 /* this sets the global state */
	geostate->load();
#endif

	/* first copy the data from static memory to shared memory */
	verts = (pfVec3 *)pfMalloc(8*sizeof(pfVec3));
	memcpy(verts, cube_vertices, 8*sizeof(pfVec3));
	norms = (pfVec3 *)pfMalloc(6*sizeof(pfVec3));
	memcpy(norms, cube_normals, 6*sizeof(pfVec3));
	lengths = (int *)pfMalloc(8*sizeof(int));
	memcpy(lengths, poly_lengths, 8*sizeof(int));
	index = (unsigned short *)pfMalloc(24*sizeof(unsigned short));
	memcpy(index, cube_polygons, 24*sizeof(unsigned short));
	sindex = (unsigned short *)pfMalloc(6*sizeof(unsigned short));
	memcpy(sindex, sequential_index, 6*sizeof(unsigned short));

	/* make a cube using 'quad' primatives (ie. quadrahedron) */
	/* make a cube using generic polygon primatives */
	geoset = new pfGeoSet;
	geoset->setPrimType(PFGS_POLYS);
	geoset->setNumPrims(6);
	geoset->setPrimLengths(lengths);
	geoset->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts, index);
	/* indexed and non-indexed attributes can't be mixed, so need  */
	/*   to use a sequential index for the rest of the attributes. */
	geoset->setAttr(PFGS_NORMAL3, PFGS_PER_PRIM, norms, sindex);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, sindex);
	geoset->setGState(geostate);

	new_geode = new pfGeode;
	new_geode->setName(name);
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(char *name, pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, NULL);
	new_geode = new pfGeode;
	new_geode->setName(name);
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/************************************************************/
/* make a marker to indicate where an intersection occures. */
pfNode *make_marker(char *name, pfVec4 *color_vec)
{
#define SPHERE_POLYS	25
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfGeoState	*geostate;

	geostate = new pfGeoState;
	geostate->makeBasic();
	geostate->setMode(PFSTATE_ENLIGHTING, PF_ON);

	geoset = pfdNewSphere(SPHERE_POLYS, pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, NULL);
	geoset->setGState(geostate);

	new_geode = new pfGeode;
	new_geode->setName(name);
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/************************************************/
/* some sample code fragments from Paul Rajlich */
#if 0
pfGeoSet *texQuad(pfGeoState *tGState, float w, float h, float alpha) {
  pfGeoSet*  geoSet = new pfGeoSet;

  pfVec2 *texCoords = (pfVec2 *) pfMalloc(4*sizeof(pfVec2));
  texCoords[0].set(0.01f, 0.99f);
  texCoords[1].set(0.01f, 0.01f);
  texCoords[2].set(0.99f, 0.01f);
  texCoords[3].set(0.99f, 0.99f);

  pfVec3 *verts3 = (pfVec3 *) pfMalloc(4*sizeof(pfVec3));
  verts3[0].set(-w/2.0f, 0.0f, h);
  verts3[1].set(-w/2.0f, 0.0f, 0.0f);
  verts3[2].set(w/2.0f, 0.0f, 0.0f);
  verts3[3].set(w/2.0f, 0.0f, h);

  pfVec4 *color = (pfVec4 *) pfMalloc(1*sizeof(pfVec4));
  color[0].set(1.0f, 1.0f, 1.0f, alpha);

  geoSet->setPrimType(PFGS_QUADS);
  geoSet->setNumPrims(1);
  geoSet->setAttr(PFGS_COORD3, PFGS_PER_VERTEX, verts3, NULL);
  geoSet->setAttr(PFGS_COLOR4, PFGS_OVERALL, color, NULL);
  geoSet->setAttr(PFGS_TEXCOORD2, PFGS_PER_VERTEX, texCoords, NULL);
  geoSet->setGState(tGState);
//  if (alpha < 1.0f)
//    geoSet->setDrawBin(2); // render transparent objects last

  return geoSet;
}

#endif

/************************************************/

