/* ======================================================================
 *
 *  CCCCC    ++      ++   pfex1_bare.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: January 4, 2002
 * CC   CC   ++      ++   Last Modified: October 29, 2003
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a VR application program that uses   */
/*   the SGI Performer library to handle the visual rendering.         */
/*      ...                                                            */
/*                                                                     */
/* The new features of this example include:                           */
/*   - adds a set of axes to the scenegraph                            */
/*   - The basic Performer Library functions:                          */
/*      * pfInit()                                                     */
/*      * pfConfig()                                                   */
/*      * pfSync()                                                     */
/*      * pfFrame()                                                    */
/*      * pfExit()                                                     */
/*      * rendering statistics                                         */
/*   - The basic vrPF Library functions:                               */
/*	* vrConfigure()                                                */
/*	* vrStart()                                                    */
/*	* vrPfMasterChannel()                                          */
/*	* vrPfPreFrame()                                               */
/*	* vrPfPostFrame()                                              */
/*	* vrExit()                                                     */
/*   - Some vrPF functions/variables used for interaction:             */
/*	* vrGet2switchValue()                                          */
/*                                                                     */
/* Functionality of this example:                                      */
/*   - Creates a single Scene Graph of a simple world, similar to the  */
/*      OpenGL example world.                                          */
/*   - Exits the application when the 'Escape' key is pressed          */
/*                                                                     */
/* Copyright 2003, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

/* Performer library includes */
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */

/* VR library includes */
#define PERFORMER
#include "freevr.h"

/* This is very important -- this function will be used as an argument. */
	void	exit_application(int);

/* functions defined in this file */
	void	update_world();
static	pfScene	*createScene();


/**********************************************************************/
int main(int argc, char **argv)
{
	/* initialize Performer */
	pfInit();

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */

	/* configure Performer */
	pfConfig();

	/* create the channels for viewing the world */
	vrStart();			/* in FreeVR, this is also where the processes are forked */

	/********************/
	/* create the world */

	/* Create a scene and assign it to the channels */
	vrPfMasterChannel()->setScene(createScene());


	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (
		!vrGet2switchValue(0)
		) {
		pfSync();
		/* latency critical operations go here */
		vrPfPreFrame();
		pfFrame();
		vrPfPostFrame();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	vrExit();
	pfExit();		/* pfBug: this actually terminates the application */
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world()
{
	/* Nothing happens -- this is a static world. */

	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(float r, float g, float b);
pfNode *make_pyramid(float r, float g, float b);
pfNode *make_axis();

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene()
{
	pfScene		*scene;
	pfGeoState	*gstate;
	pfSCS		*cs_node;
	pfMatrix	cs_matrix;

	scene = new pfScene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/**********************************************************/
	/***  Create an SCS (w/ matrix) for each static object  ***/

	/* a axis set */
	cs_matrix.makeTrans(0.0,  3.5, 0.5);
	cs_node = new pfSCS(cs_matrix);
	cs_node->addChild(make_axis());
	scene->addChild(cs_node);

	/* a blue cube */
	cs_matrix.makeTrans(-2.0,  5.5, 0.5);
	cs_node = new pfSCS(cs_matrix);
	cs_node->addChild(make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));
	scene->addChild(cs_node);

	/* a red pyramid */
	cs_matrix.makeTrans( 2.0,  5.5, 0.0);
	cs_node = new pfSCS(cs_matrix);
	cs_node->addChild(make_pyramid(255.0/256.0, 100.0/256.0, 100.0/256.0));
	scene->addChild(cs_node);

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	/* NOTE, classes/structures that are specificed by the pr library     */
	/*   do not have "new" overloaded to use shared memory for allocating */
	/*   the memory, so we need to be specifically allocate them via      */
	/*   Performer's shared memory allocation API.                        */
	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_axis()
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;
	pfMatrix	ztox_mat;
	pfMatrix	ztoy_mat;
	pfMatrix	ztoz_mat;

	/* make the Geode node that we will return */
	new_geode = new pfGeode;

	/* setup some other values needed to create the axes */
#if 1 /* for C++ */
	ztox_mat.makeRot( 90.0, 0.0, 1.0, 0.0);
	ztoy_mat.makeRot(-90.0, 1.0, 0.0, 0.0);
	ztoz_mat.makeRot(  0.0, 1.0, 0.0, 0.0);

	ztox_mat.preScale(2.0, 2.0, 2.0, ztox_mat);
	ztoy_mat.preScale(2.0, 2.0, 2.0, ztoy_mat);
	ztoz_mat.preScale(2.0, 2.0, 2.0, ztoz_mat);
#else /* for C */
	pfMakeRotMat(ztox_mat,  90.0, 0.0, 1.0, 0.0);
	pfMakeRotMat(ztoy_mat, -90.0, 1.0, 0.0, 0.0);
#endif

	/* X-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztox_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(1.0, 0.2, 0.2, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);

	/* Y-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztoy_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(0.2, 1.0, 0.2, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);

	/* Z-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztoz_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(0.2, 0.2, 1.0, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);


	return (pfNode *)new_geode;
}


/**************************/
/* draw_world(): does ... */
/**************************/
/*  NOTHING -- there is no application-specified world rendering in   */
/*    the Performer version of the CAVE library.  Performer does this */
/*    for us.                                                         */

