/* ======================================================================
 * 
 *  CCCCC    ++      ++   pfex3_dynamic.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: March 1, 2000
 * CC   CC   ++      ++   Last Modified: October 29, 2003
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a Performer VR application program   */
/*   that renders a dynamic, albeit simple, world.  The blue cube and  */
/*   red pyramid fly in a circle about a stationary green pyramid.     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - The basic Performer Library functions/classes:                  */
/*      * pfMalloc() -- function to allocate shared memory             */
/*        [NOTE: it is important to use this instead of vrShmemAlloc()]*/
/*      * pfGetFrameTimeStamp() -- used instead of vrCurrentSimTime()  */
/*      * pfDCS -- dynamic coordinate system node class                */
/*                                                                     */
/*   - The basic vrPF Library functions:                               */
/*      [none introduced in this example]                              */
/*                                                                     */
/*   - Some vrPF functions/variables used for interaction:             */
/*      [none introduced in this example]                              */
/*                                                                     */
/* Considerations:                                                     */
/*   - World simulation dynamics are done in the APP process (ie. the  */
/*       main process).                                                */
/*   - Information about the objects is passed via the scene-graph     */
/*                                                                     */
/* Copyright 2003, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

/* Performer library includes */
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */

/* VR library includes */
#define PERFORMER
#include "freevr.h"


/* Define a structure for holding a database of the virtual world. */
typedef struct {
		pfDCS		*obj1_dcs;	/* pointer to DCS node of object 1 */
		pfDCS		*obj2_dcs;	/* pointer to DCS node of object 2 */
	} WorldDataType;


/* This is very important -- this function will be used as an argument. */
	void	exit_application(int);

/* functions defined in this file */
	void	update_world(WorldDataType *wd);
static	pfScene	*createScene(WorldDataType *wd);


int main(int argc, char **argv)
{
	WorldDataType	*wd;

	/* TODO: I do the signal setup a little different in Pf than OpenGL -- investigate */
	signal(SIGINT, SIG_IGN);

	/* initialize Performer */
	pfInit();

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, exit_application);
	vrConfigure(&argc, argv, NULL);		/* call before pfConfig() -- also allow CLAs to affect config */

	/* configure Performer */
	pfConfig();

	/* create the channels for viewing the world */
	vrStart();			/* in FreeVR, this is also where the processes are forked */

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   at the expense of some performance.                       */
	vrPfMasterChannel()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);


	/********************/
	/* create the world */
	wd = (WorldDataType *)pfMalloc(sizeof(WorldDataType), pfGetSharedArena());

	/* Create a scene and assign it to the channels */
	vrPfMasterChannel()->setScene(createScene(wd));


	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (
		!vrGet2switchValue(0)
		) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also use callback routines within the  */
		/* scene-graph to modify the world.                        */
		update_world(wd);


		pfSync();
		/* latency critical operations go here */
		vrPfPreFrame();
		pfFrame();
		vrPfPostFrame();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
	vrExit();
	pfExit();	/* must come after vrExit() because FreeVR uses pfMemory */

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world(WorldDataType *wd)
{
	float		obj_x, obj_y, obj_z;
	float		sim_time = pfGetFrameTimeStamp();

	obj_z = 5.5;
	obj_x =  3*sin(sim_time * 2.0);
	obj_y = -3*cos(sim_time * 2.0) + 4.0;
	wd->obj1_dcs->setTrans(obj_x, obj_y, obj_z);

	obj_y = 4.0;
	obj_x = 3*cos(sim_time * 2.0);
	obj_z = 3*sin(sim_time * 2.0) + 5.5;
	wd->obj2_dcs->setTrans(obj_x, obj_y, obj_z);

	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(float r, float g, float b);
pfNode *make_pyramid(float r, float g, float b);
pfNode *make_axis();

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene(WorldDataType *wd)
{
	pfScene		*scene;
	pfGeoState	*gstate;
	pfDCS		*cs_node;

	scene = new pfScene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/**********************************************************/
	/*** Create a DCS (w/ matrix) for each *dynamic* object ***/

	/* an axis set */
	cs_node = new pfDCS();
	cs_node->setTrans(0.0, 3.5, 0.5);
	cs_node->addChild(make_axis());
	scene->addChild(cs_node);

	/* a blue cube */
	cs_node = new pfDCS();
	wd->obj1_dcs = cs_node;
	cs_node->setTrans(-2.0, 4.0, 5.5);
	cs_node->addChild(make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));
	scene->addChild(cs_node);

	/* a red pyramid */
	cs_node = new pfDCS();
	wd->obj2_dcs = cs_node;
	cs_node->setTrans( 2.0, 4.0, 5.5);
	cs_node->addChild(make_pyramid(255.0/256.0, 100.0/256.0, 100.0/256.0));
	scene->addChild(cs_node);

	/* a green upside down pyramid */
	cs_node = new pfDCS();
	cs_node->setTrans( 0.0, 4.0, 6.0);
	cs_node->setScale(1.0, 1.0, -1.0);	/* NOTE: flip in Z, not Y */
	cs_node->addChild(make_pyramid(100.0/256.0, 255.0/256.0, 100.0/256.0));
	scene->addChild(cs_node);

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	/* NOTE, classes/structures that are specificed by the pr library     */
	/*   do not have "new" overloaded to use shared memory for allocating */
	/*   the memory, so we need to be specifically allocate them via      */
	/*   Performer's shared memory allocation API.                        */
	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_axis()
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;
	pfMatrix	ztox_mat;
	pfMatrix	ztoy_mat;
	pfMatrix	ztoz_mat;

	/* make the Geode node that we will return */
	new_geode = new pfGeode;

	/* setup some other values needed to create the axes */
#if 1 /* for C++ */
	ztox_mat.makeRot( 90.0, 0.0, 1.0, 0.0);
	ztoy_mat.makeRot(-90.0, 1.0, 0.0, 0.0);
	ztoz_mat.makeRot(  0.0, 1.0, 0.0, 0.0);

	ztox_mat.preScale(2.0, 2.0, 2.0, ztox_mat);
	ztoy_mat.preScale(2.0, 2.0, 2.0, ztoy_mat);
	ztoz_mat.preScale(2.0, 2.0, 2.0, ztoz_mat);
#else /* for C */
	pfMakeRotMat(ztox_mat,  90.0, 0.0, 1.0, 0.0);
	pfMakeRotMat(ztoy_mat, -90.0, 1.0, 0.0, 0.0);
#endif

	/* X-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztox_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(1.0, 0.2, 0.2, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);

	/* Y-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztoy_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(0.2, 1.0, 0.2, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);

	/* Z-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztoz_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(0.2, 0.2, 1.0, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);


	return (pfNode *)new_geode;
}

