/* ======================================================================
 *
 *  CCCCC    ++      ++   pfex5_contact.c++
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: March 1, 2000
 * CC   CC   ++      ++   Last Modified: October 29, 2003
 *  CCCCC    ++      ++
 *
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a Performer application program that */
/*   allows the user to select (ie. highlight) an object by contacting */
/*   the object with the wand's avatar.                                */
/*                                                                     */
/* The new features of this example include:                           */
/*   - a cleaner method of adding objects to the array of objects      */
/*   - The basic vrPF Library functions:                               */
/*      - vrPointGetRWFrom6sensor()                                    */
/*      - vrPointTransformByMatrix()                                   */
/*                                                                     */
/* Copyright 2003, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

/* Performer library includes */
#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfMatrix & pfVec4 */

/* VR library includes */
#define PERFORMER
#include "freevr.h"
#define	WAND_SENSOR	1	/* TODO: this should really be based on the concept of props */


/* Define a structure for information about each individual object. */
typedef struct {
		pfDCS		*dcs;		/* position of the object (controlled by a pfDCS) */
		pfVec4		*color;		/* vector that contains values used to color the object */
	} ObjectType;

/* Define a structure for holding a database of the virtual world. */
typedef struct {
		int		object_count;
		int		selected_object;
		ObjectType	object[3];

		pfDCS		*wand_dcs;	/* pointer to the DCS the positions the wand avatar */
		pfVec4		*wand_color;
	} WorldDataType;

/* This is very important -- this function will be used as an argument. */
	void	exit_application(int);

/* functions defined in this file */
	void	update_world(WorldDataType *wd);
static	pfScene	*createScene(WorldDataType *wd);


int main(int argc, char **argv)
{
	WorldDataType	*wd;

	/* TODO: I do the signal setup a little different in Pf than OpenGL -- investigate */
	signal(SIGINT, SIG_IGN);

	/* initialize Performer */
	pfInit();

	/**********************************/
	/*** initialize FreeVR routines ***/
	/**********************************/
	signal(SIGINT, exit_application);
	vrConfigure(&argc, argv, NULL);

	/* configure Performer */
	pfConfig();

	/* create the channels for viewing the world */
	vrStart();

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   at the expense of some performance.                       */
	vrPfMasterChannel()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);


	/********************/
	/* create the world */
	wd = (WorldDataType *)pfMalloc(sizeof(WorldDataType), pfGetSharedArena());

	/* Create a scene and assign it to the channels */
	vrPfMasterChannel()->setScene(createScene(wd));


	/***************************/
	/*** do world simulation ***/
	/***************************/
	/* do the main simulation/render loop until 'Esc' key is pressed */
	while (
		!vrGet2switchValue(0)
		) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also use callback routines within the  */
		/* scene-graph to modify the world.                        */
		update_world(wd);


		pfSync();
		/* latency critical operations go here */
		vrPfPreFrame();
		pfFrame();
		vrPfPostFrame();
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
	vrExit();
	pfExit();

	exit(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world(WorldDataType *wd)
{
static	vrMatrix	*mat_neg90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X, -90.0);	/* convert from Pf to OpenGL space */
static	vrMatrix	*mat_pos90x = vrMatrixSetRotationIdNew(vrMatrixCreate(), VR_X,  90.0);	/* convert from OpenGL to Pf space  (ogl2pf) */
	vrPoint		wand_point;
#define wand_location wand_point.v
	pfMatrix	obj_mat;
	float		obj_x, obj_y, obj_z, tmp;
	int		prev_selected_object;
	int		obj;
	pfMatrix	tmpmat1, tmpmat2;		/* These are used by direct method of wand avatar positioning. */

#if 0 /* static world for now */
	float		obj_x, obj_y, obj_z;
	float		sim_time = pfGetTime();

	obj_z = 5.5;
	obj_x =  3*sin(sim_time * 2.0);
	obj_y = -3*cos(sim_time * 2.0) + 4.0;
	wd->object[0].dcs->setTrans(obj_x, obj_y, obj_z);

	obj_y = 4.0;
	obj_x = 3*cos(sim_time * 2.0);
	obj_z = 3*sin(sim_time * 2.0) + 5.5;
	wd->object[1].dcs->setTrans(obj_x, obj_y, obj_z);
#endif

	/* the wand's avatar is updated in the sim-process (aka APP process) */
	/*   NOTE: we could also do this with an APP-callback on the DCS node */

	/* calculate the wand DCS using the VR library */
	vrPfDCSTransform6sensor(wd->wand_dcs, WAND_SENSOR);

	/* rotate/scale the DCS so the (small) wand points forward */
	tmpmat1.makeRot(-90.0, 1.0, 0.0, 0.0);
	wd->wand_dcs->getMat(tmpmat2);	/* WARNING: unobvious call-by-ref */
	tmpmat2.preMult(tmpmat1);
	wd->wand_dcs->setMat(tmpmat2);	/* WARNING: unobvious call-by-ref */

	/* NOTE: in FreeVR, the scale of a pfDCS is not (yet) maintained when set by vrPfDCSTransform6sensor() */
	wd->wand_dcs->setScale(0.25, 0.25, 0.25);

	/* get the location of the wand */
	/* TODO: really this should be based on the concept of props, but nyi */
	vrPointGetRWFrom6sensor(&wand_point, WAND_SENSOR);
	vrPointTransformByMatrix(&wand_point, &wand_point, mat_pos90x);		/* adjust to put into Performer space */

	/* check for contact between the wand and all the objects */
	/* NOTE: this has a slight "quirk," in that the highest numbered */
	/*   object always becomes selected, even when one of the objects*/
	/*   was already selected.  A better UI would not change the     */
	/*   selected object while the button is held down.              */
	prev_selected_object = wd->selected_object;
	wd->selected_object = -1;
	for (obj = 0; obj < wd->object_count; obj++) {
		wd->object[obj].dcs->getMat(obj_mat);
		obj_mat.getRow(3, &obj_x, &obj_y, &obj_z, &tmp);
		if ( (wand_location[VR_X] < obj_x + 0.5)
		  && (wand_location[VR_X] > obj_x - 0.5)
		  && (wand_location[VR_Y] < obj_y + 0.5)
		  && (wand_location[VR_Y] > obj_y - 0.5)
		  && (wand_location[VR_Z] < obj_z + 0.5)
		  && (wand_location[VR_Z] > obj_z - 0.5) ) {
			wd->selected_object = obj;
		}
	}

	/* the selection changes, then make a change to the world to reflect this */
	if (prev_selected_object != wd->selected_object) {
		if (prev_selected_object != -1) {
			/* restore the previosly selected object */
			*(wd->object[prev_selected_object].color) *= 2.0;
		}
		if (wd->selected_object != -1) {
			/* enhance the currently selected object */
			*(wd->object[wd->selected_object].color) *= 0.5;

		}
	}
	prev_selected_object = wd->selected_object;

	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(pfVec4 *color_vec);
pfNode *make_pyramid(pfVec4 *color_vec);

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene(WorldDataType *wd)
{
	pfScene		*scene;
	pfGeoState	*gstate;
	pfDCS		*cs_node;
	pfVec4		*color;

	wd->object_count = 0;
	wd->selected_object = -1;

	scene = new pfScene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/**********************************************************/
	/*** Create a DCS (w/ matrix) for each *dynamic* object ***/

	/* a blue cube */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans(-2.0, 4.0, 3.5);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 100.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_cube(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a red pyramid */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans( 2.0, 4.0, 3.0);
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(255.0/256.0, 100.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;

	/* a green upside down pyramid */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	cs_node->setTrans( 0.0, 4.0, 4.0);
	cs_node->setScale(1.0, 1.0, -1.0);	/* NOTE: flip in Z, not Y */
	color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color->set(100.0/256.0, 255.0/256.0, 100.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(color));
	wd->object[wd->object_count].dcs = cs_node;
	wd->object[wd->object_count].color = color;
	wd->object_count++;


	/***************************************************/
	/*** Create nodes for each user interface object ***/

	/* the Wand -- a long cyan pyramid */
	cs_node = new pfDCS();
	scene->addChild(cs_node);
	wd->wand_dcs = cs_node;
	cs_node->setScale(0.25, 0.25, 0.25);
	/* NOTE: setting rotation and translation of this DCS will be uneffective   */
	/*   because those values will be overwritten by the Sensor transformation. */
	wd->wand_color = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	wd->wand_color->set(100.0/256.0, 255.0/256.0, 255.0/256.0, 1.0);
	cs_node->addChild(make_pyramid(wd->wand_color));

	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(pfVec4 *color_vec)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}

